# -*- coding: utf-8 -*-
import io
import json

from pynlple.data.jsonsource import FileJsonDataSource

from pandas import DataFrame
from pandas import read_csv, read_json
from pynlple.data.source import Source


class DataframeSource(Source):

    def __init__(self, dataframe):
        self.dataframe = dataframe

    def get_dataframe(self):
        return self.dataframe

    def set_dataframe(self, dataframe):
        self.dataframe = dataframe


class TsvDataframeSource(Source):

    def __init__(self, dataframe_path, separator='\t', quote=0, escape_char='\\', column_names=None, index_column_names=None, fill_na_map=None, encoding='utf-8', index_columns=None):
        self.path = dataframe_path
        self.separator = separator
        self.column_names = column_names
        self.na_map = fill_na_map
        self.encoding = encoding
        self.index_columns = index_columns
        self.index_column_names = index_column_names
        self.quote = quote
        self.escape_char = escape_char

    def get_dataframe(self):
        #TODO: Eats \r\n and spits sole \n in literal value strings instead
        if self.column_names:
            header = None
            names = self.column_names
        else:
            header = 'infer'
            names = None

        dataframe = read_csv(self.path,
                             sep=self.separator,
                             header=header,
                             names=names,
                             quoting=self.quote,
                             escapechar=self.escape_char,
                             encoding=self.encoding)
        if self.index_columns:
            dataframe.set_index(keys=self.index_columns, inplace=True)
        if self.na_map:
            for key, value in self.na_map.items():
                dataframe[key].fillna(value, inplace=True)
        print('Read: ' + str(len(dataframe.index)) + ' rows from ' + self.path)
        return dataframe

    def set_dataframe(self, dataframe):
        if self.column_names:
            names = self.column_names
        else:
            names = True

        if self.index_column_names:
            include_index = True
            index_names = self.index_column_names
        else:
            include_index = False
            index_names = None

        dataframe.to_csv(self.path,
                         sep=self.separator,
                         header=names,
                         index=include_index,
                         index_label=index_names,
                         quoting=self.quote,
                         escapechar=self.escape_char,
                         encoding=self.encoding)
        print('Write: ' + str(len(dataframe.index)) + ' rows to ' + self.path)


class JsonFileDataframeSource(Source):

    FILE_READ_METHOD = 'rt'
    FILE_WRITE_METHOD = 'wt'
    DEFAULT_ENCODING = 'utf-8'

    def __init__(self, json_file_path, fill_na_map=None, index_columns=None):
        self.json_file_path = json_file_path
        self.na_map = fill_na_map
        self.index_columns = index_columns

    def get_dataframe(self):
        with io.open(self.json_file_path, JsonFileDataframeSource.FILE_READ_METHOD, encoding=JsonFileDataframeSource.DEFAULT_ENCODING) as data_file:
            #TODO: implement fill_na_map
            df = read_json(data_file, orient='records', encoding=JsonFileDataframeSource.DEFAULT_ENCODING)
        return df

    def set_dataframe(self, dataframe):
        with io.open(self.json_file_path, JsonFileDataframeSource.FILE_WRITE_METHOD, encoding=JsonFileDataframeSource.DEFAULT_ENCODING) as data_file:
            #TODO: implement fill_na_map
            json.dump(dataframe.reset_index().to_dict(orient='records'), data_file, ensure_ascii=False, indent=1)


class JsonNullableFileDataframeSource(Source):

    DEFAULT_ENCODING = 'utf-8'

    def __init__(self, json_file_path, keys=None, fill_na_map=None, index_columns=None, default_encoding=DEFAULT_ENCODING):
        self.__source = JsonDataframeSource(FileJsonDataSource(file_path=json_file_path, encoding_str=default_encoding),
                                          keys=keys, fill_na_map=fill_na_map, index_columns=index_columns)

    def get_dataframe(self):
        return self.__source.get_dataframe()

    def set_dataframe(self, dataframe):
        self.__source.set_dataframe(dataframe)


class JsonDataframeSource(Source):

    def __init__(self, json_source, keys=None, fill_na_map=None, index_columns=None):
        self.json_source = json_source
        self.keys = keys
        self.na_map = fill_na_map
        self.index_columns = index_columns

    def get_dataframe(self):
        extracted_entries = list()
        for json_object in self.json_source.get_data():
            entry = dict()
            if self.keys:
                for key in self.keys:
                    if key not in json_object:
                        entry[key] = self.na_map[key]
                    else:
                        entry[key] = json_object[key]
            else:
                for key in json_object:
                    entry[key] = json_object[key]
                if self.na_map:
                    for key, value in self.na_map:
                        if key not in entry:
                            entry[key] = value
            extracted_entries.append(entry)
        dataframe = DataFrame(extracted_entries)
        if self.index_columns:
            dataframe.set_index(keys=self.index_columns, inplace=True)
        if self.na_map:
            for key, value in self.na_map.items():
                dataframe.loc[:,key].fillna(value, inplace=True)
        print('Read: ' + str(len(dataframe.index)) + ' rows from jsonsource')
        return dataframe

    def set_dataframe(self, dataframe):
        entries = dataframe.reset_index().to_dict(orient='records')
        for entry in entries:
            if self.keys:
                for key in list(entry.keys()):
                    if key not in self.keys:
                        entry.pop(key, None)
            if self.na_map:
                for key, value in self.na_map:
                    if key not in entry:
                        entry[key] = value
        self.json_source.set_data(entries)



# -*- coding: utf-8 -*-
from IPython.core.magic import Magics, magics_class, line_magic, cell_magic, line_cell_magic
from IPython.lib.backgroundjobs import BackgroundJobManager
from IPython import get_ipython
import time


@magics_class
class ThreadedJobsMagics(Magics):

    def __init__(self, shell_):
        self.jobs = BackgroundJobManager()
        super(ThreadedJobsMagics, self).__init__(shell_)

    @line_cell_magic
    def job(self, line, cell=None):
        """Run a job in new thread. May be run in cell or line fashion."""
        return self.jobs.new(cell if cell else line, self.shell.user_global_ns)

    @line_magic
    def manager(self, line=None):
        """Get a jobs manager."""
        return self.jobs

    @line_magic
    def try_kill_job(self, line):
        """Attemp to stop a job by killing the job's thread.
        Provide with a job thread object only in the line expression.
        Uses pynlple.jobs.try_kill_thread_job method."""
        thread = eval(line)
        return try_kill_thread_job(thread)


shell = get_ipython()

threaded_jobs_magics = ThreadedJobsMagics(shell)
shell.register_magics(threaded_jobs_magics)


def try_kill_thread_job(thread):
    """
    Attempts to kill a thread in which a certain job was run. Does not assures the job is stopped.
    Uses ctypes.pythonapy.PyThreadState_SetAsyncExc to inject SystemError in a process running.

    :param threading.Thread thread: thread that contains a job to stop.
    :raises ValueError: if ctypes.pythonapy.PyThreadState_SetAsyncExc failed when thread process was not found.
    :raises SystemError: if ctypes.pythonapy.PyThreadState_SetAsyncExc failed while nulling (zeroing) thread process id.
    """
    import ctypes

    id = thread.ident
    code = ctypes.pythonapi.PyThreadState_SetAsyncExc(
        ctypes.c_long(id),
        ctypes.py_object(SystemError)
    )
    if code == 0:
        raise ValueError('invalid thread id')
    elif code != 1:
        ctypes.pythonapi.PyThreadState_SetAsyncExc(
            ctypes.c_long(id),
            ctypes.c_long(0)
        )
        raise SystemError('PyThreadState_SetAsyncExc failed')


def load_ipython_extension(ipython):
    threaded_jobs_magics = ThreadedJobsMagics(ipython)
    ipython.register_magics(threaded_jobs_magics)


def unload_ipython_extension(ipython):
    pass



# -*- coding: utf-8 -*-
import logging

from pickle import PicklingError
from sklearn.externals.joblib import dump as joblib_dump
from sklearn.externals.joblib import load as joblib_load
from numpy.random import randint


logger = logging.getLogger(__name__)


def dump(explainer, filepath, *args, **kwargs):
    try:
        return joblib_dump(explainer, filepath, *args, **kwargs),
    except PicklingError as e:
        if 'Can\'t pickle <function' in str(e) and 'Explainer.__init__.<locals>.kernel' in str(e):
            logger.warning('Omitting explainer.base.kernel_fn due to: %s.\
                                            Returning the kernel_fn as the second argument' % (str(e),))
            kernel_fn = explainer.base.kernel_fn
            explainer.base.kernel_fn = None
            return joblib_dump(explainer, filepath, *args, **kwargs), kernel_fn
        else:
            raise e


def load(filepath, rollback_kernel_fn=None, rollback_kernel_width=None, *args, **kwargs):
    explainer = joblib_load(filepath, *args, **kwargs)
    if rollback_kernel_fn is not None:
        logger.warning('Applying the provided kernel_fn to the pickled explainer. %s' % (str(rollback_kernel_fn),))
        explainer.base.kernel_fn = rollback_kernel_fn
    elif explainer.base.kernel_fn is None:
        logger.warning('Applying the default kernel_fn to the pickled explainer. Kernel width: %s' % (str(rollback_kernel_width),))
        if hasattr(explainer, 'scaler'):
            stub_data = randint(0, 1, (3, explainer.scaler.mean_.shape[0]))
            if rollback_kernel_width:
                stub_explainer = explainer.__class__(stub_data, kernel_width=rollback_kernel_width)
            else:
                stub_explainer = explainer.__class__(stub_data)
        else:
            if rollback_kernel_width:
                stub_explainer = explainer.__class__(kernel_width=rollback_kernel_width)
            else:
                stub_explainer = explainer.__class__()

        explainer.base.kernel_fn = stub_explainer.base.kernel_fn
    return explainer


# -*- coding: utf-8 -*-
import io
import json
import requests
from operator import itemgetter
from elasticsearch import Elasticsearch

from pynlple.data.source import Source, SkippingSource, TakingSource
from pynlple.exceptions import DataSourceException


class ElasticJsonDataSource(TakingSource, SkippingSource):

    def __init__(self, url_addresses, port, index, query, keys=None, fill_na_map=None, authentication=None, cluster_timeout=10, max_retries=5, max_urls_pool_size=3):
        if type(url_addresses) not in (tuple, list, set):
            url_addresses = [url_addresses]
        self.urls = url_addresses
        self.PORT = port
        self.index = index
        self.__query = query
        self.keys = keys
        self.na_map = fill_na_map
        self.auth = authentication
        self.cluster_timeout = cluster_timeout
        self.max_retries = max_retries
        self.max_connection_pool_size = max_urls_pool_size
        self.es = Elasticsearch(self.__get_connection_props(), timeout=self.cluster_timeout, max_retries=max_retries, retry_on_timeout=True)
        self.es.cluster.health(wait_for_status='yellow', request_timeout=self.cluster_timeout)
        super().__init__()

    def __get_connection_props(self):
        import random
        return [{'host': url, 'port': self.PORT} for url in random.sample(self.urls, self.max_connection_pool_size)]

    def set_cluster_timeout(self, cluster_timeout_secs, max_retries=5):
        self.cluster_timeout = cluster_timeout_secs
        self.max_retries = max_retries
        self.es = Elasticsearch(self.__get_connection_props(), timeout=self.cluster_timeout, max_retries=max_retries, retry_on_timeout=True)
        self.es.cluster.health(wait_for_status='yellow', request_timeout=self.cluster_timeout)
        return self

    def _take(self, take):
        query = dict(self.__query)
        query['from'] = take
        self.__query = query

    def _skip(self, skip):
        query = dict(self.__query)
        query['size'] = skip
        self.__query = query

    def __prepare_final_query(self):
        final_query = dict(self.__query)
        if self.keys:
            final_query.update({'_source': self.keys})
        return final_query

    def get_data(self):
        try:
            final_query = self.__prepare_final_query()
            response = self.es.search(index=self.index, body=final_query)
        except Exception as e:
            raise DataSourceException(e.__str__())

        if self.na_map:
            results = [self.__fill_na_values(entry['_source']) for entry in response['hits']['hits']]
        else:
            results = [entry['_source'] for entry in response['hits']['hits']]
        print('Server had {0} total elements. Query yielded {1} elements.'
              .format(str(response['hits']['total']), str(len(results))))
        return results

    def count(self):
        try:
            final_query = self.__prepare_final_query()
            response = self.es.count(index=self.index, body=final_query)
        except Exception as e:
            raise DataSourceException(e.__str__())
        return response['count']

    def aggregate(self):
        try:
            final_query = self.__prepare_final_query()
            response = self.es.search(index=self.index, body=final_query)
        except Exception as e:
            raise DataSourceException(e.__str__())
        return response['aggregations']

    def __fill_na_values(self, entry):
        for key, value in self.na_map.items():
            if key not in entry:
                entry[key] = value
        return entry

    def set_data(self, json_data):
        raise NotImplementedError()


class ServerJsonDataSource(Source):
    """Class for providing json data via POST rest api calls."""

    def __init__(self, url_address, query, authentication=None):
        self.url_address = url_address
        self.query = query
        self.authentication = authentication

    def get_data(self):
        request = requests.post(url=self.url_address, data=self.query, auth=self.authentication)
        if request.status_code is not 200:
            raise DataSourceException('Could not reach the data. HTTP response code: {}, response: {}'.format(str(request.status_code), str(request.content)))
        else:
            return request.json()

    def set_data(self, json_data):
        raise NotImplementedError()


class FileJsonDataSource(Source):
    """Class for providing json data from json files."""

    FILE_OPEN_METHOD = 'rt'
    FILE_WRITE_METHOD = 'wt'
    DEFAULT_ENCODING = 'utf-8'

    def __init__(self, file_path, encoding_str=DEFAULT_ENCODING):
        self.file_path = file_path
        self.encoding_str = encoding_str

    def get_data(self):
        # TODO: Handle Nan/None here
        with io.open(self.file_path, FileJsonDataSource.FILE_OPEN_METHOD, encoding=self.encoding_str) as data_file:
            return json.load(data_file)

    def set_data(self, json_data):
        # TODO: Handle Nan/None here
        with io.open(self.file_path, FileJsonDataSource.FILE_WRITE_METHOD, encoding=self.encoding_str) as data_file:
            json.dump(json_data, data_file, ensure_ascii=False, indent=1)

    def __iter__(self):
        for entry in self.get_data():
            yield entry


class ESFileDumpSource(Source):

    FILE_OPEN_METHOD = 'rt'
    FILE_WRITE_METHOD = 'wt'
    DEFAULT_ENCODING = 'utf-8'

    def __init__(self, file_path, encoding_str=DEFAULT_ENCODING):
        self.file_path = file_path
        self.encoding_str = encoding_str

    def get_data(self):
        with io.open(self.file_path, FileJsonDataSource.FILE_OPEN_METHOD, encoding=self.encoding_str) as data_file:
            return list(map(itemgetter('_source'), filter(lambda x: x['_type'] == 'mention', map(json.loads, data_file))))

    def set_data(self, json_data):
        raise NotImplementedError()

    def __iter__(self):
        for entry in self.get_data():
            yield entry



# -*- coding: utf-8 -*-


class NLException(Exception):

    def __init__(self, message):
        message = '[' + self.__class__.__name__ + ']: ' + message
        super().__init__(message)

class TokenizationException(NLException):
    pass

class DataSourceException(NLException):
    pass



from collections import defaultdict
import numpy as np
import hashlib

from sklearn.base import TransformerMixin, BaseEstimator
from sklearn.utils.validation import check_is_fitted
from sklearn.utils import column_or_1d
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.cluster import KMeans
from sklearn.preprocessing import LabelBinarizer, LabelEncoder


class SafeLabelEncoder(LabelEncoder):
    """An extension of LabelEncoder that will
    not throw an exception for unseen data, but will
    instead return a default value of 99999

    Taken from https://www.alkaline-ml.com/skutil/rsts/codebase/skutil_preprocessing.html#skutil.preprocessing.SafeLabelEncoder

    Parameters
    ----------

    unseen : int,
        The integer label to be used for unseen classes. You can also use negative values (for some classifiers).
        Default: int(99999).

    Attributes
    ----------

    classes_ : the classes that are encoded
    """

    def __init__(self, unseen=99999):
        self.unseen = unseen

    def transform(self, y):
        """Perform encoding if already fit.

        Parameters
        ----------

        y : array_like, shape=(n_samples,)
            The array to encode

        Returns
        -------

        e : array_like, shape=(n_samples,)
            The encoded array
        """
        check_is_fitted(self, 'classes_')
        y = column_or_1d(y, warn=True)

        classes = np.unique(y)

        # Check not too many:
        if 0 <= self.unseen <= len(classes):
            raise ValueError('Too many factor levels in feature. Max is %i: Increase the unseen label if possible.' % self.unseen)

        e = np.array([
                         np.searchsorted(self.classes_, x) if x in self.classes_ else self.unseen
                         for x in y
                         ])

        return e


class PipeLineLabelBinarizer(TransformerMixin):

    def __init__(self, *args, **kwargs):
        self.encoder = LabelBinarizer(*args, **kwargs)

    def fit(self, X, y=None):
        self.encoder.fit(X)
        return self

    def transform(self, X, y=None):
        return self.encoder.transform(X)


class PipeLineLabelEncoder(BaseEstimator, TransformerMixin):

    def __init__(self, safe_label_encoder=False, unseen_label=-1):
        self.safe_label_encoder = safe_label_encoder
        self.unseen_label = unseen_label
        if self.safe_label_encoder:
            self.encoder = SafeLabelEncoder(unseen=self.unseen_label)
        else:
            self.encoder = LabelEncoder()

    def fit(self, X, y=None):
        self.encoder.fit(X)
        return self

    def transform(self, X, y=None):
        return self.encoder.transform(X)

    def get_params(self, deep=False):
        return {'safe_label_encoder': self.safe_label_encoder, 'unseen_label': self.unseen_label}


class VectorToCentroidCosineTransformer(object):

    def __init__(self, centroids):
        self.centroids = centroids

    def fit(self, X, y=None):
        return self

    def partial_fit(self, X, y=None):
        return self

    def transform(self, X):
        print('Calculating embedding distances to centroids {}x{}.'.format(str(len(X)), str(len(self.centroids))))
        import warnings
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=DeprecationWarning)
            return cosine_similarity(X, self.centroids)


class VectorValueThresholdReducer(object):

    def __init__(self, reduce_to=0.0):
        self.replacement_value = reduce_to
        super().__init__()

    def fit(self, X, y=None):
        self.thres_ = X.mean(axis=0) + (X.std(axis=0) * 2)
        return self

    def transform(self, X):
        X[X < self.thres_] = self.replacement_value
        return X


class Senses(object):

    def __init__(self, clusterer=KMeans()):
        self.clusterer = clusterer
        self.sense_clusters = self.clusterer.n_clusters

    def fit(self, X, y=None):
        element_index = X[:, 0]
        vector_index = X[:, 1]
        self.lexicon_size_ = element_index.shape[0]
        self.clusterer_ = self.clusterer.fit(vector_index)
        vector_distances = self.clusterer_.transform(vector_index)
        K = 50
        cluster_representatives = []
        for cluster in vector_distances.T:
            topK_indeces = np.argpartition(cluster, K)[:K]
            cluster_representatives.append(element_index[topK_indeces[np.argsort(cluster[topK_indeces])]])
        self.sense_representatives_ = np.array(cluster_representatives)
        return self


class GensimVectorModel(object):

    def __init__(self, gensim_w2v_model_path, unknown=None, keep_in_memory=False):
        self.gensim_w2v_path = gensim_w2v_model_path
        self.unknown = unknown
        self.keep = keep_in_memory
        if self.keep:
            self.load_model()

    def loaded(self):
        return hasattr(self, 'model_') and self.model_

    def load_model(self):
        print('Loading gensim w2v model.')
        from gensim.models.word2vec import Word2Vec
        self.model_ = Word2Vec.load(self.gensim_w2v_path)

    def inited(self):
        return self.loaded() and hasattr(self.model_.wv, 'syn0norm') and self.model_.wv.syn0norm is not None

    def init_model(self):
        print('Normalizing w2v model. Additional(online) fitting is now frozen.')
        self.model_.wv.init_sims(replace=True)

    def release_model(self):
        if self.loaded():
            self.model_ = None
            del self.model_
            import gc
            gc.collect()

    def __train_gensim_w2v(self, X, update):
        if self.loaded():
            print('Training loaded model. NOTE: Model could have already been in use.')
        else:
            self.load_model()
        if self.inited():
            print('WARNING: Model was inited previously. Cannot train.')
            return

        print('{}itting gensim w2v model.'.format('Ref' if update else 'F'))
        self.model_.scan_vocab(X, progress_per=25000, trim_rule=None, update=update)
        self.model_.scale_vocab(keep_raw_vocab=False, trim_rule=None, update=update)
        self.model_.finalize_vocab(update=update)

        print('Updating w2v model.')
        self.model_.train(X)

    def fit(self, X, y=None):
        self.__train_gensim_w2v(X, False)
        return self

    def partial_fit(self, X, y=None):
        self.__train_gensim_w2v(X, True)
        return self

    def __build_word2index(self):
        self.word2index = {word: idx for idx, word in enumerate(self.model_.wv.index2word)}

    def __calc_stub_vector(self):
        if not self.unknown or self.unknown is None:
            self.stub_vec_ = None
        elif self.unknown == 'mean':
            self.stub_vec_ = np.mean(np.array(self.model_.wv.syn0norm), axis=0)
        elif self.unknown == 'zero':
            self.stub_vec_ = np.zeros_like(next(iter(self.model_.wv.syn0norm)))
        else:
            self.stub_vec_ = None

    def prep(self):
        if not self.loaded():
            self.load_model()
            self.init_model()
        elif not self.inited():
            self.init_model()
        self.__build_word2index()
        self.__calc_stub_vector()

    def __get_vectors_array(self, X):
        return np.array([
            np.array([self.__get_vec_indexed(element) for element in elements])
                     for elements in X
        ])

    def __get_vec_indexed(self, element):
        try:
            return self.model_.wv.syn0norm[self.word2index[element]]
        except KeyError:
            return self.stub_vec_

    def transform(self, X):
        self.prep()
        vecs = self.__get_vectors_array(X)
        if not self.keep:
            self.release_model()
        return vecs


class Element2SenseVector(object):

    def __init__(self, sense_vectors, element2embedding, sense_threshold=None, unknown=None, override=False, fit_embeddings=False):
        self.sense_converter = VectorToCentroidCosineTransformer(sense_vectors)
        self.element2embedding = element2embedding
        self.sense_threshold = sense_threshold
        self.unknown = unknown
        self.override = override
        self.fit_embeddings = fit_embeddings

    def __calc_stub_vector(self):
        if not self.unknown or self.unknown is None:
            self.stub_vec_ = None
        elif self.unknown == 'mean':
            self.stub_vec_ = np.mean(np.array(self.token2sense_vec_.values()), axis=0)
            if self.sense_threshold:
                self.stub_vec_ = self.sense_threshold_.transform(self.stub_vec_)
        elif self.unknown == 'zero':
            self.stub_vec_ = np.zeros_like(next(iter(self.token2sense_vec_.values())))
        else:
            self.stub_vec_ = None

    def fit(self, X, y=None):
        if self.fit_embeddings:
            print('Fitting embeddings.')
            self.element2embedding_ = self.element2embedding.fit(X)
        else:
            self.element2embedding_ = self.element2embedding

        # Get embeddings for known elements.
        target_lexicon = list(set(token for tokens in X for token in tokens))
        print('Collecting known embeddings for requested elements.')

        elements_, vectors_ = zip(*filter(lambda x: x[1] is not None, zip(target_lexicon, self.element2embedding_.transform(np.array([target_lexicon]))[0])))

        print('\t{} out of requested {} element embeddings are present in inner embedding dictionary.'
              .format(str(len(elements_)), str(len(target_lexicon))))

        print('Converting embeddings to senses.')
        self.sense_converter_ = self.sense_converter.fit(vectors_)
        sense_vectors_ = self.sense_converter_.transform(vectors_)

        print('Building sense index.')
        self.token2sense_vec_ = dict(zip(elements_, sense_vectors_))

        if self.sense_threshold:
            print('Fitting threshold.')
            self.sense_threshold_ = self.sense_threshold.fit(sense_vectors_)

        self.__calc_stub_vector()
        return self

    def partial_fit(self, X, y=None):
        # Get embeddings for known elements.
        if self.fit_embeddings:
            print('Partially fitting embeddings.')
            self.element2embedding_ = self.element2embedding_.partial_fit(X)

        target_lexicon = list(set(token for tokens in X for token in tokens))

        if not self.override:
            target_lexicon = list(filter(lambda x: x not in self.token2sense_vec_, target_lexicon))

        print('[{}]: Collecting senses for requested elements with{} override.'.format(self.__class__.__name__, '' if self.override else 'out'))
        elements_and_vectors = list(filter(lambda x: x[1] is not None,
                                           zip(target_lexicon, self.element2embedding_.transform([target_lexicon])[0])))

        if len(elements_and_vectors) > 0:
            print('\t{} out of requested {} element embeddings are present in inner embedding dictionary.'
                  .format(str(len(elements_and_vectors)), str(len(target_lexicon))))
            elements_, vectors_ = zip(*elements_and_vectors)

            print('Converting embeddings to senses.')
            self.sense_converter_.partial_fit(vectors_)
            sense_vectors_ = self.sense_converter_.transform(vectors_)

            print('Updating token to sense mapping.')
            self.token2sense_vec_.update(zip(elements_, sense_vectors_))

            if self.sense_threshold:
                print('Refitting threshold.')
                self.sense_threshold_ = self.sense_threshold_.fit(np.array(list(self.token2sense_vec_.values())))

            self.__calc_stub_vector()
        else:
            print('No new elements can be fitted from {} of the input lexicon.'.format(len(target_lexicon)))
        return self

    def transform(self, X):
        if self.sense_threshold:
            return np.array([
                                np.array([self.__get_thresholded_vec_indexed(token) for token in tokens])
                                    for tokens in X
                                ])
        else:
            return np.array([
                                np.array([self.__get_vec_indexed(token) for token in tokens])
                                    for tokens in X
                                ])

    def __get_vec_indexed(self, element):
        try:
            return self.sense_threshold_.transform(self.token2sense_vec_[element])
        except KeyError:
            return self.stub_vec_

    def __get_thresholded_vec_indexed(self, element):
        try:
            return self.token2sense_vec_[element]
        except KeyError:
            return self.stub_vec_


class SenseTextTransformer(object):

    def __init__(self, element2sense, embedding_accumulator):
        self.element2sense = element2sense
        self.embeddings_accumulator = embedding_accumulator
        super().__init__()

    def fit(self, X, y=None):
        self.element2sense.fit(X)
        return self

    def partial_fit(self, X, y=None):
        self.element2sense.partial_fit(X)
        return self

    def transform(self, X):
        return np.array([self.__collect_text_embedding(embeddings) for embeddings in self.element2sense.transform(X)])

    def __collect_text_embedding(self, token_embeddings):
        known_vecs = np.array(list(filter(lambda x: x is not None, token_embeddings)))
        if len(known_vecs) <= 0:
            return self.element2sense.stub_vec
        else:
            return self.embeddings_accumulator.transform(known_vecs)


class EmbeddingAccumulator(object):
    """
    """

    def __init__(self, mean=False):
        self.get_mean = mean

    def transform(self, X):
        return self.__calc_embedding(X)

    def __calc_embedding(self, embeddings):
        if self.get_mean:
            return np.mean(embeddings, axis=0)
        else:
            return np.sum(embeddings, axis=0)


class MeanEmbeddingAccumulator(object):
    """
    This class duplicates MeanEmbeddingVectorizer, but for the functionality of normalizing the vectors.
    Also, it may be a little optimized for fully-oov-sentences.
    """

    def __init__(self, embeddings, element_index, mean_for_unknown=False):
        self.element2embedding = {element: np.array(embedding) for element, embedding in zip(element_index, embeddings)}
        self.get_mean = mean_for_unknown
        if self.get_mean:
            self.stub_vec = np.mean(embeddings, axis=0)
        else:
            self.stub_vec = np.zeros_like(self.element2embedding[next(iter(self.element2embedding))])

    def __calc_vecs_for_known_elements(self, tokens):
        if len(tokens) <= 0:
            return self.stub_vec
        elif any(token in self.element2embedding for token in tokens):
            return np.mean([self.element2embedding[token] for token in tokens if token in self.element2embedding], axis=0)
        else:
            return self.stub_vec

    def transform(self, X):
        vecs = np.array([self.__calc_vecs_for_known_elements(tokens) for tokens in X])
        return vecs


class Vector2VectorTransformer(dict):

    def __init__(self, vector_transformer):
        self.transformer = vector_transformer
        super().__init__()

    def __get_alias(self, vec):
        return hashlib.sha1(vec.view(np.uint8)).hexdigest()

    def fit(self, X, y=None):
        vectors2 = self.transformer.transform(X)
        print('Collecting transformed vectors.')
        self.vec2vec_ = dict(map(lambda vec, vec2: (self.__get_alias(vec), vec2), X, vectors2))
        return self

    def partial_fit(self, X, y=None):
        print('Finding new embedding entries.')
        new_alias2vec = {}
        for vec in X:
            alias = self.__get_alias(vec)
            if alias not in self.vec2vec_ and alias not in new_alias2vec:
                new_alias2vec[alias] = vec

        # Transform to sense vectors
        print('Calculating embedding distances to senses.')
        aliases, vecs = zip(*new_alias2vec.items())
        transformed_vecs = self.transformer.transform(np.array(vecs))

        # Add to tokenVec2senseVec dictionary
        print('Collecting sense vectors.')
        self.vec2vec_.update(zip(aliases, transformed_vecs))
        return self

    def transform(self, X):
        aliases = [self.__get_alias(vec) for vec in X]
        return np.array([self.vec2vec_[vec_alias] if vec_alias in self.vec2vec_ else None
                         for vec_alias in aliases])


class TfidfEmbeddingVectorizer(object):
    def __init__(self, word2vec):
        self.word2vec = word2vec
        self.word2weight_ = None
        self.dim = len(next(iter(word2vec.items()))[1])
        self.empty_stub = np.zeros(self.dim)

    def fit(self, X, y=None):
        tfidf = TfidfVectorizer(analyzer=lambda x: x)
        tfidf.fit(X)
        # if a word was never seen - it must be at least as infrequent
        # as any of the known words - so the default idf is the max of
        # known idf's
        max_idf = max(tfidf.idf_)
        self.word2weight_ = defaultdict(
            lambda: max_idf,
            [(w, tfidf.idf_[i]) for w, i in tfidf.vocabulary_.items()])

        return self

    def transform(self, X):
        return np.array([
                np.mean([self.word2vec[w] * self.word2weight_[w]
                         for w in words if w in self.word2vec] or
                        [self.empty_stub], axis=0)
                for words in X
            ])


# -*- coding: utf-8 -*-
from pynlple.data.source import Source
from pynlple.exceptions import DataSourceException
from pynlple.module import is_folder, is_file, list_dir, append_paths


class FilePathSource(Source):
    """Class for providing filepaths from data folders."""

    def __init__(self, paths, extension_suffix=None):
        self.paths = paths
        self.extension = extension_suffix

    def get_files(self):
        accumulated_paths = list()
        for path in self.paths:
            if is_file(path):
                accumulated_paths.append(path)
            elif is_folder(path):
                files = list_dir(path)
                if self.extension:
                    for file in filter(lambda f: f.endswith(self.extension), files):
                        accumulated_paths.append(append_paths(path, file))
                else:
                    for file in files:
                        accumulated_paths.append(append_paths(path, file))
            else:
                raise DataSourceException('Path {0} does not exist/is neither file nor folder!'.format(path))
        return accumulated_paths

    def __iter__(self):
        for file_ in self.get_files():
            yield file_



# -*- coding: utf-8 -*-
import unittest

from pynlple.module import abs_path

from pynlple.processing.dictionary import FileFolderTokenMapping
from pynlple.processing.stopwords import DictionaryStopwordsFilter


class StopwordsFilterTest(unittest.TestCase):

    def test_should_init_with_default_params(self):
        self.filter = DictionaryStopwordsFilter(FileFolderTokenMapping('punctuation.txt', data_folder=abs_path(__file__, 'data/stopwords')))
        self.assertIsNotNone(self.filter)

    def test_should_filter_with_provided_dictionary(self):
        self.filter = DictionaryStopwordsFilter(FileFolderTokenMapping('punctuation.txt', data_folder=abs_path(__file__, 'data/stopwords')))
        tokens = 'These are some tokens , we need to filter ! So , don\'t be so shy to do it .'.split()

        expected_tokens = 'These are some tokens we need to filter So don\'t be so shy to do it'.split()
        self.assertEqual(expected_tokens, self.filter.filter(tokens))

    def test_should_filter_with_provided_dictionary_and_preprocessing_rule(self):
        preprocessing = lambda x: x if len(x) > 2 else '.'
        self.filter = DictionaryStopwordsFilter(FileFolderTokenMapping('punctuation.txt', data_folder=abs_path(__file__, 'data/stopwords')),
                                                prep=preprocessing)
        tokens = 'These are some tokens , we need to filter ! So , don\'t be so shy to do it .'.split()

        expected_tokens = 'These are some tokens need filter don\'t shy'.split()
        self.assertEqual(expected_tokens, self.filter.filter(tokens))



# -*- coding: utf-8 -*-
import re
import html
import unicodedata

from itertools import chain
from functools import partial
from operator import itemgetter

from .emojis import UnicodeClusteredEmojiProviderV11
from .token import TOKEN_TYPES, DELIMITER_TYPES, DELIMITER_CONTEXTS, \
    EMOJIS, PUNCTUATION, SPECIAL_SYMBOLS, CURRENCY, TOKEN_DEFAULT_DELIM


URL_REGEX = r'((ht|f)tps?:[\\/]+|www\d{0,3}\.)([.…]+|[^\s\\/$.?#].[^\s]*)'


class IPreprocessor(object):
    """Interface to collect modules for text string preprocessing."""

    def preprocess(self, string_):
        return None


class Replacer(IPreprocessor):
    """Preprocessor interface for modules that replace certain text entities with normalized tags, etc."""


class StackingPreprocessor(IPreprocessor):
    """Class for continuous stacking and applying preprocessors in a natural order."""

    def __init__(self, preprocessor_list=None):
        if not preprocessor_list:
            self.preprocessors = []
        else:
            self.preprocessors = list(preprocessor_list)

    def append_preprocessor(self, preprocessor):
        self.preprocessors.append(preprocessor)

    def prepend_preprocessor(self, preprocessor):
        self.preprocessors.insert(0, preprocessor)

    def insert_preprocessor(self, position, preprocessor):
        self.preprocessors.insert(position, preprocessor)

    def preprocess(self, string_):
        out_string = string_
        for preprocessor in self.preprocessors:
            out_string = preprocessor.preprocess(out_string)
        return out_string


class ToLowerer(IPreprocessor):
    """Lowers all alpha characters."""

    def preprocess(self, string_):
        return string_.lower()


class CharacterUnicodeCategoryReplacer(IPreprocessor):

    WHITESPACE = ' '
    CURRENCY_SIGN = '$'
    NEWLINE = '\n'
    decimals = set(map(str, (range(10))))

    def preprocess(self, string_):
        return ''.join(filter(None, map(self.get_code, string_)))

    def get_code(self, char):
        if char in self.decimals:
            return char
        elif char in ['\r', '\n']:
            return char

        cat = unicodedata.category(char)
        if cat in ['Zs']:
            return CharacterUnicodeCategoryReplacer.WHITESPACE
        if cat in ['Zp', 'Zl']:
            return CharacterUnicodeCategoryReplacer.NEWLINE
        elif cat in ['Sc']:
            return CharacterUnicodeCategoryReplacer.CURRENCY_SIGN
        elif cat in ['Lu', 'Lt']:
            return cat[0].upper()
        elif cat[0] in ['P', 'p']:
            return char
        elif cat in ['Mc', 'Mn']:
            return None
        else:
            return cat[0].lower()


class SoftTextScissors(Replacer):
    """Cuts the end of the string in a 'soft' manner: the first whitespace after the 'soft_length' is
    sought for in reach of 'window' size rightwards. If no whitespace is present in the window, then
    the 'soft_length' of string is taken."""

    def __init__(self, soft_length=1000, window=50, whitespace_regex=r'\s'):
        self.soft_length = soft_length
        self.window = window
        self.whitespace_regex = whitespace_regex
        self.__ws_patt = re.compile(self.whitespace_regex)
        self.max_length = self.soft_length + self.window
        self.min_length = self.soft_length - self.window

    def preprocess(self, string_):
        return _find_soft_edges(string_, self.soft_length, self.min_length, self.max_length, self.__ws_patt)


def _find_soft_edges(string_, soft_length, min_length, max_length, ws_patt):
    if len(string_) > soft_length:
        ws = ws_patt.search(string_, pos=soft_length, endpos=max_length)
        if ws:
            return string_[:ws.start()]
        else:
            r_ws = ws_patt.search(string_[min_length:soft_length][::-1])
            if r_ws:
                return string_[:soft_length-r_ws.end()]
            else:
                return string_[:soft_length]
    else:
        return string_


class HeadNTailTextScissors(Replacer):
    """Cuts out the middle of the string in a 'soft' manner."""

    def __init__(self, soft_length=1000, window=50, whitespace_regex=r'\s', connector=' ... '):
        self.soft_length = soft_length
        self.window = window
        self.whitespace_regex = whitespace_regex
        self.connector = connector
        self.__ws_patt = re.compile(self.whitespace_regex)
        self.max_length = self.soft_length + self.window + len(self.connector)
        self.__half_pos = int(self.soft_length / 2)
        self.__half_window = int(self.window / 2)
        self.__half_window_max_pos = self.__half_pos + self.__half_window + 1
        self.__half_window_min_pos = self.__half_pos - self.__half_window - 1

    def preprocess(self, string_):
        if len(string_) > self.max_length:
            l_string_ = _find_soft_edges(string_, self.__half_pos, self.__half_window_min_pos, self.__half_window_max_pos, self.__ws_patt)
            r_string_ = _find_soft_edges(string_[::-1], self.__half_pos, self.__half_window_min_pos, self.__half_window_max_pos, self.__ws_patt)[::-1]
            return l_string_ + self.connector + r_string_
            # l_ws = self.__ws_patt.search(string_, pos=self.__half_pos, endpos=self.__half_window_max_pos)
            # l_ = l_ws.start() if l_ws else self.__half_pos
            # r_ws = self.__ws_patt.search(string_[::-1], pos=self.__half_pos, endpos=self.__half_window_max_pos)
            # r_ = r_ws.start() if r_ws else self.__half_pos
            # return string_[:l_] + self.connector + string_[-r_:]
        else:
            return string_


class RegexReplacer(Replacer):
    """Preprocessor. Replaces entities described by a specific regular expression from a text string
    with a stated string. Supports \n groups usage in target string."""

    def __init__(self, regex_query_string, group_string_repl_or_method, case_sensitive, use_locale, restrict_to_ascii):
        self.target = group_string_repl_or_method
        flag_1 = re.IGNORECASE if not case_sensitive else 0
        flag_2 = re.LOCALE if use_locale else 0
        flag_3 = re.ASCII if restrict_to_ascii else 0
        self.pattern = re.compile(regex_query_string, flag_1 | flag_2 | flag_3)
        super().__init__()

    def preprocess(self, string_):
        return self.pattern.sub(self.target, string_)


class RegexReplacerAdapter(RegexReplacer):

    def __init__(self, regex_query, replace_tag_with=None, case_sensitive=False, use_locale=False, restrict_to_ascii=False):
        if not replace_tag_with:
            replace_tag_with = ''
        super().__init__(regex_query, replace_tag_with, case_sensitive, use_locale, restrict_to_ascii)


def _get_bold_replacement(repl_):
    stripd_repl_ = repl_.strip()
    if len(stripd_repl_) == 0:
        return repl_, repl_, repl_, repl_
    else:
        ind_ = repl_.index(stripd_repl_)
        ind_end_ = ind_ + len(stripd_repl_)
        return (repl_,
                ''.join([repl_[:ind_], '<b>', repl_[ind_:]]),
                ''.join([repl_[:ind_end_], '</b>', repl_[ind_end_:]]),
                ''.join([repl_[:ind_], '<b>', repl_[ind_:ind_end_], '</b>', repl_[ind_end_:]]))


bold_tag_pattern = re.compile(r'</?b>')


def replace_w_bold_tag_preserved(repl_, left_bound_repl_, right_bound_repl_, left_right_bound_repl_, match):
    match_text = match.group(0)
    if '<b>' in match_text or '</b>' in match_text:
        cnt = 0
        for match in bold_tag_pattern.finditer(match_text):
            if match.group(0) == '</b>':
                cnt -= 1
            else:
                cnt += 1
        if cnt > 0:
            return left_bound_repl_
        elif cnt < 0:
            return right_bound_repl_
        else:
            return left_right_bound_repl_
    else:
        return repl_


def _get_target_func(repl_):
    return partial(replace_w_bold_tag_preserved, *_get_bold_replacement(repl_))


class BoldPreservingRegexReplacerAdapter(RegexReplacerAdapter):

    def __init__(self, regex_query, replace_tag_with=None, case_sensitive=False, use_locale=False, restrict_to_ascii=False, preserve_bold=False):
        if not preserve_bold:
            target = replace_tag_with
        else:
            target = _get_target_func(replace_tag_with)
        super().__init__(regex_query, target, case_sensitive, use_locale, restrict_to_ascii)


class MultiWhitespaceReplacer(RegexReplacerAdapter):
    """Replace numerous whitespaces (\s+) in an input string with a default single space ' '."""

    def __init__(self, replace_tag_with=' '):
        regex_query = r'\s+'
        super().__init__(regex_query, replace_tag_with)


class MultiNewLineReplacer(RegexReplacerAdapter):
    """Replace numerous newlines ([\r\n]+) in an input string with a default single newline '\r\n'."""

    def __init__(self, replace_tag_with='\r\n'):
        regex_query = r'[\r\n]+'
        super().__init__(regex_query, replace_tag_with)


class NewLineSpaceReplacer(RegexReplacerAdapter):
    """Replace numerous newlines ([\r\n]+) in an input string with a default single space ' '."""

    def __init__(self, replace_tag_with=' '):
        regex_query = r'[\r\n]+'
        super().__init__(regex_query, replace_tag_with)


class MultiInLineWhitespaceReplacer(RegexReplacerAdapter):
    """Replace numerous in-line whitespaces ([\f   \t\v]+) in an input string with a default single space ' '.
    Note: there are 3 types of spaces as represented here: https://en.wikipedia.org/wiki/Punctuation"""

    def __init__(self, replace_tag_with=' '):
        regex_query = r'[\f   \t\v]+'
        super().__init__(regex_query, replace_tag_with)


class Trimmer(Replacer):
    """Trims (or strips) heading and trailing whitespaces."""

    def preprocess(self, string_):
        return string_.strip()


class HtmlTagReplacer(RegexReplacerAdapter):
    """Replaces all tags of format <tag> and </tag> including tags with attributes and inner values: <a href='..'>."""

    def __init__(self, replace_tag_with=None):
        regex_query = r'<.*?>'
        super().__init__(regex_query, replace_tag_with)


class VKMentionReplacer(RegexReplacerAdapter):
    """Replaces the inner VK links and user mention of type '[<id>|<name>]'."""

    def __init__(self, replace_tag_with=None):
        regex_query = r'\[[\w_\-:]+\|.*?\]'
        super().__init__(regex_query, replace_tag_with, False, False, False)


class AtReferenceReplacer(RegexReplacerAdapter):
    """Replaces the inner VK links and user mention of type '@<id>' (or '@<name>'?).
    This replacer finds only such entities which are separate words (start the line or
    have blank whitespace)"""

    def __init__(self, replace_tag_with=None):
        regex_query = r'((?<=\s)|(?<=^))@[\w_-]+'
        super().__init__(regex_query, replace_tag_with, False, False, False)


class URLReplacer(BoldPreservingRegexReplacerAdapter):
    """Research https://mathiasbynens.be/demo/url-regex and https://gist.github.com/dperini/729294
    Since we need to increase recall of link-a-like entities in text, we adopt the algorithm with
    some possible erroneous cases.
    I used regexp from https://mathiasbynens.be/demo/url-regex by @stephenhay
    and extended with ftps, www, wwwDDD prefixes, and "\" variant of slash ("/")"""

    def __init__(self, replace_tag_with=None, preserve_bold_tag=False):
        super().__init__(URL_REGEX, replace_tag_with, False, False, False, preserve_bold_tag)


class EmailReplacer(RegexReplacerAdapter):
    """Replaces all e-mail-like entities, non-latin including. Includes length check for each entity part."""

    def __init__(self, replace_tag_with=None):
        regex_query = r'[\w0-9][\w0-9._%+-]{0,63}@(?:[\w0-9](?:[\w0-9-]{0,62}[\w0-9])?\.){1,8}[\w]{2,63}'
        super().__init__(regex_query, replace_tag_with, False, False, False)


class UserWroteRuReplacer(RegexReplacerAdapter):
    """Replaces automatically generated sequences of citing found on forums. Grammar belike: <user> wrote <date?>:"""
# ((\d\d\s\w\w\w\s\d\d\d\d|\w+)(,\s\d\d:\d\d)?)?
    def __init__(self, replace_tag_with=None):
        regex_query = r'[^\s]+\s(\(\d\d\.\d\d\.\d\d\s\d\d:\d\d\)\s)?писал\((а|a)\)\s?((\d\d\s\w\w\w\s\d\d\d\d|\w+)(,\s\d\d:\d\d)?\s?)?:'
        super().__init__(regex_query, replace_tag_with, False, False, False)


class DigitsReplacer(RegexReplacerAdapter):
    """Replaces all series of digits."""

    def __init__(self, multi, replace_tag_with=None):
        regex_query = r'\d+' if multi else r'\d'
        super().__init__(regex_query, replace_tag_with, False)


class CommaReplacer(RegexReplacerAdapter):
    """Replaces variations of comma taken from https://en.wikipedia.org/wiki/Punctuation.
    Symbols replaced: [,،、，] to the default ','. Please, ref to http://unicode-table.com to decode symbols."""

    DEFAULT_REPLACEMENT = ","

    def __init__(self, replace_tag_with=DEFAULT_REPLACEMENT):
        regex_query = r'[,،、，]'
        super().__init__(regex_query, replace_tag_with, False, False, False)


class QuotesReplacer(RegexReplacerAdapter):
    """Replaces variations of single qoute (actually there are not much of them,
    but there are some look-alike symbols from the top of the unicode table
    that resemble single quotes and may be used in its function.
    So, take care when using this class. Symbols replaced: ['`ʹʻʼʽ՚‘’‚‛′‵] to the
    default [']. Please, ref to http://unicode-table.com to decode symbols."""

    DEFAULT_REPLACEMENT = '\''

    def __init__(self, replace_tag_with=DEFAULT_REPLACEMENT):
        # Some of the symbols are taken from other than punctuation.txt section
        regex_query = r'[\'`ʹʻʼʽ՚‘’‚‛′‵]'
        # This section does not contain quotes from not-punctuation.txt section
        # regex_query = r'[\'‘’‚‛′‵]'
        super().__init__(regex_query, replace_tag_with, False, False, False)


class DoubleQuotesReplacer(RegexReplacerAdapter):
    """Replaces variations of double qoute (actually there are not much of them,
    but there are some look-alike symbols from the top of the unicode table
    that resemble double quotes and may be used in its function.
    So, take care when using this class. Symbols replaced: ["«»ʺ“”„‟″‶] to the
    default [\"]. Please, ref to http://unicode-table.com to decode symbols."""

    DEFAULT_REPLACEMENT = "\""

    def __init__(self, replace_tag_with=DEFAULT_REPLACEMENT):
        regex_query = r'["«»ʺ“”„‟″‶]'
        super().__init__(regex_query, replace_tag_with, False, False, False)


class DoubleQuotesEscaper(RegexReplacerAdapter):
    """Escapes the double qoutes symbol with the corresponding string tag"""

    DEFAULT_REPLACEMENT = "DQTS"

    def __init__(self, replace_tag_with=DEFAULT_REPLACEMENT):
        regex_query = r'\"'
        super().__init__(regex_query, replace_tag_with, False, False, False)


class DashAndMinusReplacer(RegexReplacerAdapter):
    """Replaces variations of dash/minus (actually there are not much of them,
    but there are some look-alike symbols from the top of the unicode table
    that resemble dash and minus and may be used in its function.
    So, take care when using this class. Symbols replaced: [-‐‑‒–—―－−] to the
    default [-]. Please, ref to http://unicode-table.com to decode symbols."""

    def __init__(self, replace_tag_with='-'):
        regex_query = r'[-‐‑‒–—―－−]'
        super().__init__(regex_query, replace_tag_with, False, False, False)


class SoftHyphenReplacer(RegexReplacerAdapter):
    """Removes soft hyphen."""

    def __init__(self, replace_tag_with=''):
        regex_query = r'[­]'
        super().__init__(regex_query, replace_tag_with, False)


class TripledotReplacer(RegexReplacerAdapter):
    """Replaces triple dot one-symbol expression with 3 separate fullstops."""

    def __init__(self, replace_tag_with='...'):
        regex_query = r'[…]'
        super().__init__(regex_query, replace_tag_with, False)


class OpenParenthesisReplacer(RegexReplacerAdapter):

    def __init__(self, replace_tag_with='('):
        regex_query = r'[（]'
        super().__init__(regex_query, replace_tag_with, False)


class CloseParenthesisReplacer(RegexReplacerAdapter):

    def __init__(self, replace_tag_with=')'):
        regex_query = r'[）]'
        super().__init__(regex_query, replace_tag_with, False)


class QuestionMarkReplacer(RegexReplacerAdapter):

    def __init__(self, replace_tag_with='?'):
        regex_query = r'[？]'
        super().__init__(regex_query, replace_tag_with, False)


class ExclamationMarkReplacer(RegexReplacerAdapter):

    def __init__(self, replace_tag_with='!'):
        regex_query = r'[！]'
        super().__init__(regex_query, replace_tag_with, False)


class BoldTagReplacer(RegexReplacerAdapter):

    def __init__(self, replace_tag_with=''):
        regex_query = r'</?b>'
        super().__init__(regex_query, replace_tag_with, False)


class MultiPunctuationReplacer(RegexReplacerAdapter):
    """Replaces all multiple punctuation and special symbols (default keyboard set + some additional:
    ref to data/punctuation.txt, data/special_symbols.txt, data/currency.txt) with single corresponding symbol."""

    def __init__(self, replace_tag_with='\\1', max_punctuation_allowed=2, reduce_to=2):
        self.max = max_punctuation_allowed
        self.leave = reduce_to
        regex_query = r'([' + ''.join(map(re.escape, chain(PUNCTUATION, SPECIAL_SYMBOLS, CURRENCY))) + r'])\1{' + str(self.max) + ',}'
        super().__init__(regex_query, replace_tag_with * self.leave, False, False, False)


class MultiLetterReplacer(RegexReplacerAdapter):
    """Replaces multiple consecutive letters ([^\W\d_]) with a number letters of this type (thus preserving possible
    doubling/tripling of letters but reducing possible exhaustive lettterrring."""

    def __init__(self, replace_tag_with='\\1', max_letters_allowed=3, reduce_to=3):
        self.max = max_letters_allowed
        self.leave = reduce_to
        regex_query = r'([^\W\d_])\1{' + str(self.max) + ',}'
        super().__init__(regex_query, replace_tag_with * self.leave, False, False, False)


class NonWhitespaceAlphaNumPuncSpecSymbolsAllUnicodeReplacer(RegexReplacerAdapter):
    """Removes all unicode NON alpha-numeric, punctuation and special symbols (default keyboard set + some additional:
    ref to data/punctuation.txt, data/special_symbols.txt, data/currency.txt), whitespaces and emojis."""

    def __init__(self, replace_tag_with=''):
        regex_query = r'[^\s\w' + ''.join(map(re.escape, chain(PUNCTUATION, SPECIAL_SYMBOLS, CURRENCY, EMOJIS))) + ']'
        super().__init__(regex_query, replace_tag_with, False, False, False)


class NonWordOrNumberOrWhitespaceAllUnicodeReplacer(RegexReplacerAdapter):
    """This cannot handle _ correctly. This also cannot handle soft hyphen correctly."""

    def __init__(self, replace_tag_with=' '):
        parts = [
            r'(?<=\d)[,:\.](?!\d)|(?<!\d)[,:\.](?=\d)|(?<!\d)[,:\.](?!\d)',  # this leaves out digital sequences with , : . in middle
            r'(?<=[\w\d])[\'\-](?![\w\d])|(?<![\w\d])[\'\-](?=[\w\d])|(?<![\w\d])[\'\-](?![\w\d])',  # this leaves out alphanumeric tokens with - and ' in middle
            r'[^\s\w\d\'\-,:\.]+',  # all other non-whitespace, non-alpha and non-digits
            r'_',
            ]
        regex_query = r'(' + r'|'.join(parts) + r')+'
        super().__init__(regex_query, replace_tag_with, False, False, False)


class WordTokenizer(RegexReplacer):
    """"""

    PRESPLIT_PAT = re.compile(r' +')

    def __init__(self, token_delim=TOKEN_DEFAULT_DELIM):
        # token_types = TOKEN_TYPES
        self.token_regex = r'|'.join(TOKEN_TYPES)
        self.delimiter = token_delim
        self.pattern = re.compile(self.token_regex)

    def __tokenize_substr(self, str_):
        for matches in self.pattern.finditer(str_):
            span = matches.span()
            yield matches.string[span[0]:span[1]]

    def preprocess(self, string_):
        return self.delimiter.join(chain.from_iterable(map(self.__tokenize_substr, WordTokenizer.PRESPLIT_PAT.split(string_))))


class MinimalWordTokenizer(IPreprocessor):
    """"""

    PRESPLIT_PAT = re.compile(r' +')

    def __init__(self, token_delim=TOKEN_DEFAULT_DELIM):
        regex = r'|'.join(TOKEN_TYPES[:-1])
        self.pattern = re.compile(regex)
        self.delimiter = token_delim

    def __find_tokens(self, str_):
        offset = 0
        length = len(str_)
        for matches in self.pattern.finditer(str_):
            span = matches.span()
            start = span[0]
            if offset != start:
                for c in str_[offset:start]:
                    yield c
            end = span[1]
            offset = end
            yield str_[start:end]
        if offset != length:
            for c in str_[offset:length]:
                yield c

    def preprocess(self, string_):
        return self.delimiter.join(chain.from_iterable(map(self.__find_tokens, self.PRESPLIT_PAT.split(string_))))


class DelimiterTokenizer(IPreprocessor):
    """"""

    TOKEN_DEFAULT_DELIM = ' '
    PRESPLIT_PAT = re.compile(r' +')

    def __init__(self, token_delim=TOKEN_DEFAULT_DELIM):
        delims = DELIMITER_TYPES
        regex_query = r'(' + '|'.join(delims) + ')'
        self.pattern = re.compile(regex_query)
        self.delimiter = token_delim

    def preprocess(self, string_):
        return self.delimiter.join(
            chain.from_iterable(
                map(
                    partial(filter, None),
                    map(self.pattern.split, self.PRESPLIT_PAT.split(string_))
                )
            )
        )


class DelimiterFindingTokenizer(IPreprocessor):

    PRESPLIT_PAT = re.compile(r' +')

    def __init__(self, token_delim=TOKEN_DEFAULT_DELIM):
        delims = DELIMITER_CONTEXTS
        regex_query = r'|'.join(delims)
        self.pattern = re.compile(regex_query)
        self.delim=token_delim

    def __find_tokens(self, str_):
        splits = {0, len(str_)}
        for matches in self.pattern.finditer(str_):
            for i in range(1, len(matches.regs)):
                span = matches.span(i)
                if span != (-1,-1):
                    splits.add(span[0])
                    splits.add(span[1])
        splits = list(sorted(splits))
        for s, e in zip(splits[:-1], splits[1:]):
            yield str_[s:e]

    def preprocess(self, string_):
        return self.delim.join(
            chain.from_iterable(
                map(
                    partial(filter, None),
                    map(self.__find_tokens, self.PRESPLIT_PAT.split(string_))
                )
            )
        )


class HtmlEntitiesEscaper(Replacer):

    def preprocess(self, string_):
        return html.escape(string_)


class HtmlEntitiesUnescaper(Replacer):

    def preprocess(self, string_):
        return html.unescape(string_)


class PhoneNumberReplacer(RegexReplacerAdapter):
    """Replaces the phone-number-looking entities."""

    def __init__(self, replace_tag_with=' phoneNumberTag '):
        regex_query = r'(?<!\d)(\+?\d{1,3})?[\(\- ]+\d{3,4}[\)\- ]+\d{2,3}[\- ]?[\d\- ]{4}\d(?!\d)'
        super().__init__(regex_query, replace_tag_with, False, False, False)


class PriceReplacer(RegexReplacerAdapter):
    """Replaces the money entities with a corresponding tag. Encircles the name of the currency."""

    def __init__(self, replace_tag_with=' moneyAmountTag '):
        regex_query = r'[\d.]+(\s)?(?:р|грн|руб|рублей|рубля|EUR|' + '|'.join(map(re.escape, CURRENCY)) + r')\s*[\s.]'
        super().__init__(regex_query, replace_tag_with, False, False, False)


class EmojiReplacer(RegexReplacerAdapter):
    """Replaces emojis (from unicode table) with the corresponding tag.
    Emoji list is taken from https://github.com/carpedm20/emoji"""
    def __init__(self, replace_tag_with=' emojiTag '):
        super().__init__(r'(' + '|'.join(map(re.escape, EMOJIS)) + r')', replace_tag_with, False, False, False)


class EmojiClusterReplacer(RegexReplacerAdapter):
    """Replaces emojis (from unicode table) with the corresponding tag and cluster name.
    Note the name of the cluster is a number."""

    def __init__(self, tag_prefix=' emojiTag '):
        self.raw_tag_prefix = tag_prefix
        self.tag_prefix = tag_prefix.strip()
        s = tag_prefix.index(self.tag_prefix)
        e = s + len(self.tag_prefix)

        self.emoji_tags = dict((emoji, self.raw_tag_prefix[:s] + cluster + self.raw_tag_prefix[e:]) for emoji, cluster in UnicodeClusteredEmojiProviderV11(self.tag_prefix))

        def select_tag(match):
            match_text = match.group(0)
            return self.emoji_tags.get(match_text, self.raw_tag_prefix)

        super().__init__(r'(' + '|'.join(map(re.escape, map(itemgetter(0), self.emoji_tags.items()))) + r')', select_tag)


class DefaultPreprocessorStack(StackingPreprocessor):
    def __init__(self, preserve_newlines=True, soft_cut=None):
        self.preserve_nl = preserve_newlines
        stack = [
            HtmlEntitiesUnescaper(),
            BoldTagReplacer(),
            HtmlTagReplacer(' '),
            URLReplacer(' urlTag '),
            VKMentionReplacer(' vkMentionTag '),
            EmailReplacer(' emailTag '),
            PhoneNumberReplacer(' phoneNumberTag '),
            AtReferenceReplacer(' atRefTag '),
            UserWroteRuReplacer(' userWroteRuTag '),
            CommaReplacer(),
            QuotesReplacer(),
            DoubleQuotesReplacer(),
            SoftHyphenReplacer(),
            DashAndMinusReplacer(),
            OpenParenthesisReplacer(),
            CloseParenthesisReplacer(),
            QuestionMarkReplacer(),
            ExclamationMarkReplacer(),
            TripledotReplacer(),
            NonWhitespaceAlphaNumPuncSpecSymbolsAllUnicodeReplacer(''),
            ToLowerer(),
            DigitsReplacer(False, '0'),
            MultiPunctuationReplacer(),
            MultiLetterReplacer(),
            MultiNewLineReplacer(),
            MultiInLineWhitespaceReplacer() if self.preserve_nl else MultiWhitespaceReplacer(),
            WordTokenizer(),
            Trimmer(),
            # Any other?
        ]
        if soft_cut:
            stack.insert(3, soft_cut)
        super().__init__(stack)


class ExtendedReplacerStack(DefaultPreprocessorStack):
    """
    Creates an extended (by appending to the end) instance of preprocessor.DefaultPreprocessorStack.

    @:param list list_extension_replacers: a list of preplacers to extend the default set with
    """
    def __init__(self, list_extension_replacers):
        super().__init__()
        self.extension_replacers = list_extension_replacers
        for replacer in self.extension_replacers:
            super().append_preprocessor(replacer)




# -*- coding: utf-8 -*-
import io

from pynlple.exceptions import DataSourceException
from pynlple.module import append_paths, list_dir, is_file, strip_extension, abs_path, exists, is_folder


class DictionaryLookUp(object):
    """
    Dictionary class to store tokens sequences with their classes.
    Implements tree-like storage structure. Supports preprocessing
    of both saved and query sequences.

    Constructor takes:
    1. An iterator of `(sequence, tag)` tuples, where
    sequence can be either a whitespace-separated `str` or `list(str)` of tokens,
    2. A `func(list(str)) -> list(str)` for preprocessing of both contained (before saving)
    and queried (at the moment of query) sequences that is applied
    to each 'str' token in a sequence (`str.split()` first).

    Use built-in `lookup.__setitem__(str_or_list_str, tag_str)` or `lookup.put(str_or_list_str, tag_str)`
    to add new sequence-class mappings. Mappings are overriden if the sequence mapping
    is present in the lookup dictionary.

    Use built-in `lookup.__getitem__(str_or_list_str)` or `lookup.get_tag(str_or_list_str)` to query your
    sequence and get either the class assigned tag or `None` if no corresponding mapping can be found.

    Usage::

        >>> lookup = DictionaryLookUp([('sequence one', 'class1'), (['sequence', 'two'], 'class2')], str.lower)
        >>> lookup['Sequence One']
        class1
        >>> lookup['sequence two'] = 'class1'
        >>> lookup[['sequence' 'two']]
        class1
        >>> lookup[['sequence three']]
        None
    """

    __X = '@endX!'
    """
    Internal tag to mark sequence end. Please, make sure to not have this in your sequences.

    :type str
    """

    def __init__(self, token_mappings, preprocessing_method=None):
        """
        Inits a dictionary lookup from mappings of tokens to their class tags.
        Includes the usafe of a preprocessing function.

        :param iterator(tuple(list(str), object)) token_mappings: iterable of tuples
        of token sequences and corresponding tags
        :param func(list(str)) -> list(str) preprocessing_method: the preprocessing function to be applied
        on stored sequences and queried sequences. Default: None, no preprocessing
        """
        self.dictionary = dict()
        self.preprocess = preprocessing_method
        for tokens, class_ in token_mappings:
            self.__setitem__(tokens, class_)

    def __setitem__(self, tokens, class_):
        if type(tokens) == str:
            tokens = tokens.split()
        DictionaryLookUp.__add_to_dict(self.__prep(tokens), 0, self.dictionary, class_)

    def __getitem__(self, tokens):
        if type(tokens) == str:
            tokens = tokens.split()
        return DictionaryLookUp.__get_from_dict(self.__prep(tokens), 0, self.dictionary)

    @staticmethod
    def __add_to_dict(list_, index, dict_, tag):
        if index < len(list_):
            if list_[index] not in dict_:
                dict_[list_[index]] = dict()
            DictionaryLookUp.__add_to_dict(list_, index + 1, dict_[list_[index]], tag)
        elif index == len(list_):
            dict_.update({DictionaryLookUp.__X: tag})

    @staticmethod
    def __get_from_dict(list_, index, dict_):
        if index < len(list_):
            if list_[index] not in dict_:
                return None
            return DictionaryLookUp.__get_from_dict(list_, index + 1, dict_[list_[index]])
        elif index == len(list_):
            if DictionaryLookUp.__X in dict_:
                return dict_[DictionaryLookUp.__X]
            else:
                return None

    def put(self, tokens, class_):
        """
        Add new or override the existing token sequence - class mapping to the lookup.

        :param list(str) tokens: sequence of tokens to set mapping for.
        Also supports a `str` sequence by applying str.split()
        :param object class_: corresponding class tag
        :return: void
        """
        if type(tokens) == str:
            tokens = tokens.split()
        DictionaryLookUp.__add_to_dict(self.__prep(tokens), 0, self.dictionary, class_)

    def get_tag(self, tokens):
        """
        Query the token sequence to infer its mapped class tag.

        :param list(str) tokens: sequence of tokens to get mapping for.
        Also supports a `str` sequence by applying str.split()
        :return: the corresponding class tag mapping or `None` if such sequence does not contain mapping
        :rtype: object
        """
        if type(tokens) == str or type(tokens) == list:
            tokens = tokens.split()
        return DictionaryLookUp.__get_from_dict(self.__prep(tokens), 0, self.dictionary)

    def __prep(self, tokens):
        if self.preprocess:
            return [self.preprocess(token) for token in tokens]
        else:
            return tokens

    def get_longest_sequence(self):
        """
        Get the longest stored token sequence length.

        :return: the length of the longest sequence
        :rtype: int
        """
        return DictionaryLookUp.__find_local_max(self.dictionary, 0, 0)

    def get_known_classes(self):
        """
        Get the set of classes known by the current lookup.

        :return: a set of string class names
        :rtype: set
        """
        return set(DictionaryLookUp.__find_all_classes(self.dictionary, []))

    @staticmethod
    def __find_local_max(dict_, current, max_):
        if current > max_:
            max_ = current
        for key in dict_.keys():
            if key != DictionaryLookUp.__X:
                max_ = DictionaryLookUp.__find_local_max(dict_[key], current + 1, max_)
        return max_

    @staticmethod
    def __find_all_classes(dict_, classes_list):
        for key in dict_.keys():
            if key == DictionaryLookUp.__X:
                classes_list.append(dict_[key])
            else:
                classes_list = DictionaryLookUp.__find_all_classes(dict_[key], classes_list)
        return classes_list


class FileFolderTokenMapping(object):
    """
    Class instance searches over the file-folder tree and implements `iterator(tuple(str, str))`
    (via `__iter__()`) over all text files surpassing specific conditions yielding `(file-line, file-path-name)`
    tuple pairs consumable by :class:DictionaryLookUp.

    File-folder tree walk is performed relative to `data_folder`, which is by default the 'data' folder
    in plynlple.processing package. Tree walk may be `recursive` (visiting inner subfolders),
    `extension`-sensitive (checking file suffixes to match the `extension` if not None).

    Tree walk is performed for all files and folders in the source `data_folder` (including
    `recursive` and `extension` rules) unless any subpaths are stated in `source_paths'.
    All 'source_paths' are treated exclusively to all other datafile paths. These `source_paths` also
    undergo `recursive` and `extension` rule checks.

    The whole relative path (to the data path) to the file is put as the second item in the output tuple.
    The extension of the file is stripped, though.

    You can mix up `list`/`set`/`tuple`s of file paths, folder paths as well as
    `dictionary`s of `{prefix:suffixes} in `source_paths`.

    Consider the following folder structure::
        -folder1
            -folder1-1
            -file1-1 (folder1/file1-1)
            -file1-2 (folder1/file1-2)
        -folder2
            -folder2-1
                -file2-1-1 (folder2/folder2-1/file2-1-1)
                -file2-1-2 (folder2/folder2-1/file2-1-2)
            -folder2-2
                -file2-2-1 (folder2/folder2-2/file2-2-1)
            -file2-1 (folder2/file2-1)
        -file1 (file1)
        -file2 (file2)

    Setting the `source_paths` to ['folder1', 'folder2/folder2-1] will yield the following tree walked::
        -folder1
            -folder1-1
            -file1-1 (folder1/file1-1)
            -file1-2 (folder1/file1-2)
        -folder2
            -folder2-1
                -file2-1-1 (folder2/folder2-1/file2-1-1)
                -file2-1-2 (folder2/folder2-1/file2-1-2)

    Complex `source_paths` may look like following:
    {'folder2': ['folder2-1', 'folder2-2']}, ['folder1/file1-1', 'file1'] and yield::
        -folder2
            -folder2-1
                -file2-1-1 (folder2/folder2-1/file2-1-1)
                -file2-1-2 (folder2/folder2-1/file2-1-2)
            -folder2-2
                -file2-2-1 (folder2/folder2-2/file2-2-1)
        -folder1
            -file1-1 (folder1/file1-1)
        -file1 (file1)

    Use this class to arrange you Named Entities, POS lists, stopwords, etc. by putting
    the corresponding files to the folders by groups (e.g. rus/stopwords/punctuation.txt)

    """

    COMMENT_SEQUENCE = '###'

    def __init__(self, source_paths=None, data_folder=None, recursive=True, extension='.txt', encoding='utf8'):
        """
        Create an `iterator(tuple(str, str))` over all text files surpassing specific conditions
        yielding `(file-line, file-path-name)` tuple pairs consumable by :class:DictionaryLookUp.

        :param iterable source_paths: paths relative to `data_folder` to walk through exclusively
        :param str data_folder: paths to the folder containing data,
        :param bool recursive: visit subfolders
        :param str extension: filter files by suffix, default: None - perform no filtering
        :param str encoding: text file encodings, default: `utf8`
        """
        if data_folder:
            self.__data_folder = data_folder
        else:
            self.__data_folder = abs_path(__file__, 'data')
        self.__recursive = recursive
        self.__extension = extension
        self.__encoding = encoding
        self.__file_tag_mappings = set()
        self.__parse_mapping_sources(self.__file_tag_mappings, '', source_paths)

    def __parse_mapping_sources(self, file_tag_mappings, path_prefix, source_paths):
        if not source_paths or len(source_paths) == 0:
            self.__sweep_folder_or_file(file_tag_mappings, path_prefix, '', True)
        elif type(source_paths) is str:
            self.__sweep_folder_or_file(file_tag_mappings, path_prefix, source_paths, True)
        elif type(source_paths) is list or type(source_paths) is set:
            for seed_path in source_paths:
                self.__parse_mapping_sources(file_tag_mappings, path_prefix, seed_path)
        elif type(source_paths) is dict and all(type(key) is str for key in source_paths):
            for prefix, suffix in source_paths.items():
                self.__parse_mapping_sources(file_tag_mappings, append_paths(path_prefix, prefix), suffix)

    def __sweep_folder_or_file(self, file_tag_mappings, path_prefix, folder_or_file_subpath, recursive):
        absolute_file_or_folder_path = append_paths(self.__data_folder, path_prefix)
        if folder_or_file_subpath and len(folder_or_file_subpath) > 0:
            absolute_file_or_folder_path = append_paths(absolute_file_or_folder_path, folder_or_file_subpath)
        if not exists(absolute_file_or_folder_path):
            if self.__extension and exists(absolute_file_or_folder_path + self.__extension):
                absolute_file_or_folder_path += self.__extension
            else:
                raise DataSourceException('[{0}] File/folder could not be found: {1}, extension: {2}'
                                          .format(__name__, absolute_file_or_folder_path, self.__extension))
        if is_file(absolute_file_or_folder_path):
            # Here check if we have extension set. Use it to check file ending if set
            if not self.__extension or absolute_file_or_folder_path.endswith(self.__extension):
                tag = FileFolderTokenMapping\
                    .normalize_tag(append_paths(path_prefix, strip_extension(folder_or_file_subpath)))
                file_tag_mappings.add((absolute_file_or_folder_path, tag))
        if is_folder(absolute_file_or_folder_path):
            if recursive:
                for element_name in list_dir(absolute_file_or_folder_path):
                    new_path_prefix = append_paths(path_prefix, folder_or_file_subpath)
                    self.__sweep_folder_or_file(file_tag_mappings, new_path_prefix, element_name, self.__recursive)

    @staticmethod
    def normalize_tag(tag):
        """
        Use to replace wrong backslashes with correct slashes.

        :param str tag: tag text derived from file path
        :return: tag with nice slashes
        :rtype: str
        """
        return tag.replace('\\', '/')

    def get_file_tag_mappings(self):
        """
        Get the file paths of files retrieved during file-folder tree walking.
        All conditions applied.

        :return: set of tuple(full-filepath, tag) mappings
        :rtype: set(str, str)
        """
        return self.__file_tag_mappings

    def __iter__(self):
        """
        Get the iterable of tuple(token_sequence, token_tag) from
        the files walked. Ignores lines with "###" at the start, enabling commenting.

        :return: tuples of text lines (token strings) and corresponding tags derived
        from the containing filepath
        :rtype: iterator(tuple(str, str))
        """
        for file_path, tag in self.__file_tag_mappings:
            with io.open(file_path, 'rt', encoding=self.__encoding) as input_file:
                for line in input_file:
                    line = line.strip()
                    if line.startswith(FileFolderTokenMapping.COMMENT_SEQUENCE):
                        continue
                    yield line.split(), tag




# -*- coding: utf-8 -*-
import unittest

from pynlple.processing.preprocessor import *


class SoftTextScissorsTest(unittest.TestCase):

    def test_should_not_malform_input(self):
        input_string = 'this is text to cut with softcut scissors'
        preprocessor = SoftTextScissors(soft_length=50, window=4)
        result = preprocessor.preprocess(input_string)
        expected = 'this is text to cut with softcut scissors'
        self.assertEqual(expected, result)

    def test_should_not_malform_input_edge_length(self):
        input_string = 'this is text to cut with softcut scissors'
        preprocessor = SoftTextScissors(soft_length=41, window=4)
        result = preprocessor.preprocess(input_string)
        expected = 'this is text to cut with softcut scissors'
        self.assertEqual(expected, result)

    def test_should_cut_whitespace_edge_length(self):
        input_string = 'this is text to cut with softcut scissors '
        preprocessor = SoftTextScissors(soft_length=41, window=4)
        result = preprocessor.preprocess(input_string)
        expected = 'this is text to cut with softcut scissors'
        self.assertEqual(expected, result)

    def test_should_not_cut_whitespace_not_edge_length(self):
        input_string = 'this is text to cut with softcut scissors '
        preprocessor = SoftTextScissors(soft_length=42, window=4)
        result = preprocessor.preprocess(input_string)
        expected = 'this is text to cut with softcut scissors '
        self.assertEqual(expected, result)

    def test_should_soft_cut_input_rightwards(self):
        input_string = 'this is text to cut with softcut scissors'
        preprocessor = SoftTextScissors(soft_length=28, window=5)
        result = preprocessor.preprocess(input_string)
        expected = 'this is text to cut with softcut'
        self.assertEqual(expected, result)

    def test_should_soft_cut_input_leftwards(self):
        input_string = 'this is text to cut with softcut scissors'
        preprocessor = SoftTextScissors(soft_length=28, window=4)
        result = preprocessor.preprocess(input_string)
        expected = 'this is text to cut with'
        self.assertEqual(expected, result)

    def test_should_soft_cut_input_not_soft(self):
        input_string = 'this is text to cut with softcut scissors'
        preprocessor = SoftTextScissors(soft_length=30, window=1)
        result = preprocessor.preprocess(input_string)
        expected = 'this is text to cut with softc'
        self.assertEqual(expected, result)

    def test_should_soft_cut_input_edge_trailing_space(self):
        input_string = 'this is text to cut with softcut scissors'
        preprocessor = SoftTextScissors(soft_length=25, window=4)
        result = preprocessor.preprocess(input_string)
        expected = 'this is text to cut with'
        self.assertEqual(expected, result)

    def test_should_soft_cut_input_edge_ending_space(self):
        input_string = 'this is text to cut with softcut scissors'
        preprocessor = SoftTextScissors(soft_length=15, window=4)
        result = preprocessor.preprocess(input_string)
        expected = 'this is text to'
        self.assertEqual(expected, result)


class HeadNTailTextScissorsTest(unittest.TestCase):

    def test_should_not_malform_input(self):
        input_string = 'this is text to cut with softcut scissors'
        preprocessor = HeadNTailTextScissors(soft_length=50, window=4)
        result = preprocessor.preprocess(input_string)
        expected = 'this is text to cut with softcut scissors'
        self.assertEqual(expected, result)

    def test_should_not_malform_input_edge_length(self):
        input_string = 'this is text to cut with softcut scissors'
        preprocessor = HeadNTailTextScissors(soft_length=41, window=4)
        result = preprocessor.preprocess(input_string)
        expected = 'this is text to cut with softcut scissors'
        self.assertEqual(expected, result)

    def test_should_not_cut_whitespace_not_edge_length(self):
        input_string = 'this is text to cut with softcut scissors '
        preprocessor = HeadNTailTextScissors(soft_length=42, window=4)
        result = preprocessor.preprocess(input_string)
        expected = 'this is text to cut with softcut scissors '
        self.assertEqual(expected, result)

    def test_should_soft_cut_input_all_rightwards(self):
        input_string = 'this is text to cut with soft scissors ya'
        preprocessor = HeadNTailTextScissors(soft_length=28, window=4)
        result = preprocessor.preprocess(input_string)
        expected = 'this is text to ... soft scissors ya'
        self.assertEqual(expected, result)

    def test_should_soft_cut_input_all_leftwards(self):
        input_string = 'this is text which is cut with softcut scissors'
        preprocessor = HeadNTailTextScissors(soft_length=28, window=4)
        result = preprocessor.preprocess(input_string)
        expected = 'this is text ... softcut scissors'
        self.assertEqual(expected, result)

    def test_should_soft_cut_input_left_and_rightwards(self):
        input_string = 'this is text which is cut with soft scissors ya'
        preprocessor = HeadNTailTextScissors(soft_length=28, window=4)
        result = preprocessor.preprocess(input_string)
        expected = 'this is text ... soft scissors ya'
        self.assertEqual(expected, result)

    def test_should_soft_cut_input_right_and_leftwards(self):
        input_string = 'this is text to cut with softcut scissors'
        preprocessor = HeadNTailTextScissors(soft_length=28, window=4)
        result = preprocessor.preprocess(input_string)
        expected = 'this is text to ... softcut scissors'
        self.assertEqual(expected, result)

    def test_should_soft_cut_input_not_soft(self):
        input_string = 'this is text which is cut with softcut scissors'
        preprocessor = HeadNTailTextScissors(soft_length=30, window=1)
        result = preprocessor.preprocess(input_string)
        expected = 'this is text wh ... oftcut scissors'
        self.assertEqual(expected, result)


class HtmlEntitiesReplacerTest(unittest.TestCase):

    def test_should_not_malform_input(self):
        input_string = 'this is some input, that does not contain html-escaped entities. this should be left untouched!'
        replacer = HtmlEntitiesEscaper()
        result = replacer.preprocess(input_string)
        expected = 'this is some input, that does not contain html-escaped entities. this should be left untouched!'
        self.assertEqual(expected, result)

    def test_should_replace_symbols_with_html_entities(self):
        input_string = 'this\'s input with "escaped" entities & some other sheet'
        replacer = HtmlEntitiesEscaper()
        result = replacer.preprocess(input_string)
        expected = 'this&#x27;s input with &quot;escaped&quot; entities &amp; some other sheet'
        self.assertEqual(expected, result)

    def test_should_replace_html_entities_with_symbols(self):
        input_string = 'this&#x27;s input with &quot;escaped&quot; entities &amp; some other sheet'
        replacer = HtmlEntitiesUnescaper()
        result = replacer.preprocess(input_string)
        expected = 'this\'s input with "escaped" entities & some other sheet'
        self.assertEqual(expected, result)


class VKMentionReplacerTest(unittest.TestCase):

    def test_should_replace_with_default(self):
        input_string = 'Здесь в тексте есть ссылка[id254351532|Олололя] как в вконтакте.'
        replacer = VKMentionReplacer()
        result = replacer.preprocess(input_string)
        expected = 'Здесь в тексте есть ссылка как в вконтакте.'
        self.assertEqual(expected, result)

    def test_should_replace_with_tag(self):
        input_string = 'Здесь в тексте есть ссылка[id254351532|Олололя] как в вконтакте.'
        replacer = VKMentionReplacer('tag')
        result = replacer.preprocess(input_string)
        expected = 'Здесь в тексте есть ссылкаtag как в вконтакте.'
        self.assertEqual(expected, result)

    def test_should_replace_hard_case(self):
        input_string = 'Здесь в тексте есть ссылка[id254351532:bp-16297716_231892|Coca-Cola] как в вконтакте.'
        replacer = VKMentionReplacer('tag')
        result = replacer.preprocess(input_string)
        expected = 'Здесь в тексте есть ссылкаtag как в вконтакте.'
        self.assertEqual(expected, result)

    def test_should_not_replace_with_square_brackets(self):
        input_string = 'Здесь в тексте нет ссылки [Олололя] как в вконтакте.'
        replacer = VKMentionReplacer('tag')
        result = replacer.preprocess(input_string)
        expected = 'Здесь в тексте нет ссылки [Олололя] как в вконтакте.'
        self.assertEqual(expected, result)

    def test_should_replace_with_tag_and_square_brackets_afterwards(self):
        input_string = 'Здесь в тексте есть ссылка[id254351532|Олололя] как в [любимом] вконтакте.'
        replacer = VKMentionReplacer('tag')
        result = replacer.preprocess(input_string)
        expected = 'Здесь в тексте есть ссылкаtag как в [любимом] вконтакте.'
        self.assertEqual(expected, result)


class RegexReplacerTest(unittest.TestCase):

    TEST_REGEX = 'a+'
    TEST_TARGET = 'b'
    TEST_STRING = 'cCaaAAaACc'

    TEST_GROUP_REGEX = '(a+)'
    TEST_GROUP_TARGET = '\\1b\\1'

    TEST_W_REGEX = '\\w'
    TEST_W_TARGET = 'w'
    TEST_W_STRING = 't+tЫ+ЫІ+Іj+j'

    TEST_W_RARE_STRING = '\u263A+⛇+ۑ+Ҝ+Θ+ᰔ+⃣a+ꋘ'

    def test_should_perform_case_insensitive_replacement(self):
        replacer = RegexReplacer(RegexReplacerTest.TEST_REGEX, RegexReplacerTest.TEST_TARGET, False, False, False)
        result = replacer.preprocess(RegexReplacerTest.TEST_STRING)
        expected = 'cCbCc'
        self.assertEqual(expected, result)

    def test_should_perform_case_sensitive_replacement(self):
        replacer = RegexReplacer(RegexReplacerTest.TEST_REGEX, RegexReplacerTest.TEST_TARGET, True, False, False)
        result = replacer.preprocess(RegexReplacerTest.TEST_STRING)
        expected = 'cCbAAbACc'
        self.assertEqual(expected, result)

    def test_should_perform_w_replacement_without_unicode_symbols(self):
        replacer = RegexReplacer(RegexReplacerTest.TEST_W_REGEX,
                                 RegexReplacerTest.TEST_W_TARGET,
                                 False, False, True)
        result = replacer.preprocess(RegexReplacerTest.TEST_W_STRING)
        expected = 'w+wЫ+ЫІ+Іw+w'
        self.assertEqual(expected, result)

    def test_should_perform_w_replacement_with_unicode_symbols(self):
        replacer = RegexReplacer(RegexReplacerTest.TEST_W_REGEX,
                                 RegexReplacerTest.TEST_W_TARGET,
                                 False, False, False)
        result = replacer.preprocess(RegexReplacerTest.TEST_W_STRING)
        expected = 'w+ww+ww+ww+w'
        self.assertEqual(expected, result)

    def test_should_perform_w_replacement_with_unicode_rare_symbols(self):
        replacer = RegexReplacer(RegexReplacerTest.TEST_W_REGEX,
                                 RegexReplacerTest.TEST_W_TARGET,
                                 False, False, False)
        result = replacer.preprocess(RegexReplacerTest.TEST_W_RARE_STRING)
        expected = '\u263A+⛇+w+w+w+w+⃣w+w'
        self.assertEqual(expected, result)

    def test_should_replace_rare_double_quotes_in_stack_with_single_quote_replacer(self):
        replacer_1 = QuotesReplacer()
        replacer_2 = DoubleQuotesReplacer()

        replacer = StackingPreprocessor([replacer_1, replacer_2])

        # First symbol is a comma, second one - is a quote
        input_string = ',‚тестируем′ «Нестле» '
        output = replacer.preprocess(input_string)

        expected = ',\'тестируем\' \"Нестле\" '
        self.assertEqual(expected, output)

    def test_should_remove_all_except_words_and_numbers(self):
        punct_string = '!!!1.word.\' \'2.word-with-hypen 3.word\'s here! ' \
                       '4.6 12:60 12,000!? -no way-^_^'
        replacer = NonWordOrNumberOrWhitespaceAllUnicodeReplacer()
        result = replacer.preprocess(punct_string)
        expected = ' 1 word   2 word-with-hypen 3 word\'s here  ' \
                   '4.6 12:60 12,000   no way '
        self.assertEqual(expected, result)

    def test_should_reduce_multiple_punctuation_symbols(self):
        punct_string = ' ... !!...!?????asddf --a !!kek shlak&&**&&&&'
        replacer = MultiPunctuationReplacer(max_punctuation_allowed=1, reduce_to=1)
        result = replacer.preprocess(punct_string)
        expected = ' . !.!?asddf -a !kek shlak&*&'
        self.assertEqual(expected, result)

    def test_should_reduce_multiple_letter_symbols(self):
        letttter_string = 'a aa aaa aaaa aaaaaa aaaaaaa aaa b aabaa aaaabaaaa'
        replacer = MultiLetterReplacer()
        result = replacer.preprocess(letttter_string)
        expected = 'a aa aaa aaa aaa aaa aaa b aabaa aaabaaa'
        self.assertEqual(expected, result)

    def test_should_reduce_multiple_letter_symbols_custom_length(self):
        letttter_string = 'a aa aaa aaaa aaaaaa aaaaaaa aaa b aabaa aaaabaaaa'
        replacer = MultiLetterReplacer(max_letters_allowed=2, reduce_to=1)
        result = replacer.preprocess(letttter_string)
        expected = 'a aa a a a a a b aabaa aba'
        self.assertEqual(expected, result)

    def test_should_reduce_multiple_letter_symbols_reduced_to_zero(self):
        letttter_string = 'a aa aaa aaaa aaaaaa aaaaaaa aaa b aabaa aaaabaaaa'
        replacer = MultiLetterReplacer(max_letters_allowed=2, reduce_to=0)
        result = replacer.preprocess(letttter_string)
        expected = 'a aa      b aabaa b'
        self.assertEqual(expected, result)

    def test_should_replace_email(self):
        email_string = 'Hi, guys! I had email@yahoo.com, I didn\'t use @ at yahoo.com. Here\'s my new e-mail: my_email+test@mail.yahoo.com. '\
                       'Can you please validate it?'
        replacer = EmailReplacer(' emailTag ')
        result = replacer.preprocess(email_string)
        expected = 'Hi, guys! I had  emailTag , I didn\'t use @ at yahoo.com. Here\'s my new e-mail:  emailTag . '\
                   'Can you please validate it?'
        self.assertEqual(expected, result)

    def test_should_replace_bold_tag(self):
        punct_string = '#<b>лотерея</b> #<b>выигрыш</b> #<b>лото</b> #<b>тираж</b> #<b>Спортлото</b>'
        replacer = BoldTagReplacer()
        result = replacer.preprocess(punct_string)
        expected = '#лотерея #выигрыш #лото #тираж #Спортлото'
        self.assertEqual(expected, result)

    def test_should_not_replace_unicode_emojis(self):
        emoji_string = 'Some smiles: \U0001F1E8\U0001F1EC, \U0001F1EC, \U0001F192\U0001F192. And digits 00000.'
        replacer = NonWhitespaceAlphaNumPuncSpecSymbolsAllUnicodeReplacer('')
        result = replacer.preprocess(emoji_string)
        expected = 'Some smiles: \U0001F1E8\U0001F1EC, \U0001F1EC, \U0001F192\U0001F192. And digits 00000.'
        self.assertEqual(expected, result)

    def test_should_replace_at_reference(self):
        ref_string = 'в предложении @giveaway_ukraine_ua2 ссылка на человека'
        replacer = AtReferenceReplacer('atRefTag')
        result = replacer.preprocess(ref_string)
        expected = 'в предложении atRefTag ссылка на человека'
        self.assertEqual(expected, result)


class TokenizerTest(unittest.TestCase):

    def test_should_tokenize_words(self):
        input_string = '!These words, un-like-ly said here(sic!), should be tokenized. ' \
                  'Ukrainska mova takoj z\'yavliaetsia tyt.'
        tokenizer = WordTokenizer()
        expected = '! These words , un-like-ly said here ( sic ! ) , should be tokenized . ' \
                   'Ukrainska mova takoj z\'yavliaetsia tyt .'

        result = tokenizer.preprocess(input_string)
        self.assertEqual(expected, result)

    def test_should_tokenize_words_w_cyrillics(self):
        input_string = 'за 2011 год). Вот полный список торговых марок от Nestle, которые можно найти в любом российском магазине: Nescafé, КитКат,'
        tokenizer = WordTokenizer()
        expected = 'за 2011 год ) . Вот полный список торговых марок от Nestle , которые можно найти в любом российском магазине : Nescafé , КитКат ,'

        result = tokenizer.preprocess(input_string)
        self.assertEqual(expected, result)

    def test_should_tokenize_digits_w_special_symbols(self):
        input_string = '0.0, 000 00-00 00- 00\' 0.00, 00.00, 00:00, 0000.00.00, 0000:00, 000,00, 00..00 :00 0'
        tokenizer = WordTokenizer()
        expected = '0.0 , 000 00 - 00 00 - 00 \' 0.00 , 00.00 , 00:00 , 0000.00.00 , 0000:00 , 000,00 , 00 .. 00 : 00 0'

        result = tokenizer.preprocess(input_string)
        self.assertEqual(expected, result)

    def test_should_tokenize_words_w_special_symbols(self):
        input_string = 'п\'ятого \'november\' з-під рову- merry-go-round'
        tokenizer = WordTokenizer()
        expected = 'п\'ятого \' november \' з-під рову - merry-go-round'

        result = tokenizer.preprocess(input_string)
        self.assertEqual(expected, result)

    def test_should_tokenize_unicode_emojis(self):
        emoji_string = 'Some smiles: \U0001F1E8, \U0001F1EC, \U0001F192\U0001F192'
        tokenizer = WordTokenizer()
        result = tokenizer.preprocess(emoji_string)
        expected = 'Some smiles : \U0001F1E8 , \U0001F1EC , \U0001F192 \U0001F192'
        self.assertEqual(expected, result)

    def test_should_tokenize_not_listed_emojis(self):
        emoji_string = 'огромное спасибо за помощь❤️❤❤️\r\nзачисление перевода'
        tokenizer = WordTokenizer()
        result = tokenizer.preprocess(emoji_string)
        expected = 'огромное спасибо за помощь ❤️ ❤ ❤️ \r\n зачисление перевода'
        self.assertEqual(expected, result)

    def test_should_not_tokenize_double_unicode_emojis(self):
        emoji_string = 'Some smiles: \U0001F1E8\U0001F1EC, \U0001F1EC, \U0001F192\U0001F192'
        tokenizer = WordTokenizer()
        result = tokenizer.preprocess(emoji_string)
        expected = 'Some smiles : \U0001F1E8\U0001F1EC , \U0001F1EC , \U0001F192 \U0001F192'
        self.assertEqual(expected, result)

    def test_should_not_tokenize_words_w_dash(self):
        input_string = 'wi-fi'
        tokenizer = WordTokenizer()
        expected = 'wi-fi'

        result = tokenizer.preprocess(input_string)
        self.assertEqual(expected, result)

    def test_should_tokenize_words_w_digits_w_special_symbols(self):
        input_string = '25.двадцатьпять 15\'16-річний я 15\'ок \'солюшнз-'
        tokenizer = WordTokenizer()
        expected = '25 . двадцатьпять 15 \' 16-річний я 15\'ок \' солюшнз -'

        result = tokenizer.preprocess(input_string)
        self.assertEqual(expected, result)

    def test_should_tokenize_words_w_digits(self):
        input_string = '25двадцатьпять 15річний15'
        tokenizer = WordTokenizer()
        expected = '25 двадцатьпять 15 річний 15'

        result = tokenizer.preprocess(input_string)
        self.assertEqual(expected, result)

    def test_should_tokenize_words_and_not_multiple_dots(self):
        input_string = '...я ... не знаю... что и... сказать'
        tokenizer = WordTokenizer()
        expected = '... я ... не знаю ... что и ... сказать'

        result = tokenizer.preprocess(input_string)
        self.assertEqual(expected, result)

    def test_should_replace_urls_with_tags_isnide(self):
        replacement = 'LINK'
        input_string = 'this is the link http://dostavka-gaza-v-<b>dom</b>.<b>ru</b> we need to replace.'
        tokenizer = URLReplacer(replace_tag_with=replacement)
        expected = 'this is the link ' + replacement + ' we need to replace.'

        result = tokenizer.preprocess(input_string)
        self.assertEqual(expected, result)

    def test_should_replace_url(self):
        replacement = 'LINK'
        input_string = 'this is the link http://dostavka-gaza-v-dom.ru we need to replace.'
        tokenizer = URLReplacer(replace_tag_with=replacement)
        expected = 'this is the link ' + replacement + ' we need to replace.'
        result = tokenizer.preprocess(input_string)
        self.assertEqual(expected, result)

    def test_should_replace_url2(self):
        replacement = 'LINK'
        input_string = 'this is the link http://www.rbc.ru/rbcfreenews/58f5f3aa9a79477a6ddcc1d3#xtor=AL-[internal_traffic]--[rss.rbc.ru]-[top_stories_brief_news] we need to replace.'
        tokenizer = URLReplacer(replace_tag_with=replacement)
        expected = 'this is the link ' + replacement + ' we need to replace.'
        result = tokenizer.preprocess(input_string)
        self.assertEqual(expected, result)


    def test_should_replace_truncated_url(self):
        replacement = 'LINK'
        input_string = 'this is the link http://... we need to replace.'
        tokenizer = URLReplacer(replace_tag_with=replacement)
        expected = 'this is the link ' + replacement + ' we need to replace.'

        result = tokenizer.preprocess(input_string)
        self.assertEqual(expected, result)

    def test_should_replace_user_wrote_reply_sign_at_start(self):
        replacement = '<tag>'
        input_string = 'Perfect4a писал(а) 27 авг 2015, 18:01: This is test'
        tokenizer = UserWroteRuReplacer(replace_tag_with=replacement)
        expected = '' + replacement + ' This is test'

        result = tokenizer.preprocess(input_string)
        self.assertEqual(expected, result)

    def test_should_replace_user_wrote_reply_sign_in_middle(self):
        replacement = '<tag>'
        input_string = 'This is test Perfect4a писал(а) 27 авг 2015, 18:01: This is test'
        tokenizer = UserWroteRuReplacer(replace_tag_with=replacement)
        expected = 'This is test ' + replacement + ' This is test'

        result = tokenizer.preprocess(input_string)
        self.assertEqual(expected, result)

    def test_should_replace_user_wrote_reply_second_type_sign_at_start(self):
        replacement = '<tag>'
        input_string = 'mircurt (03.03.16 20:39) писал(а): This is test'
        tokenizer = UserWroteRuReplacer(replace_tag_with=replacement)
        expected = '' + replacement + ' This is test'

        result = tokenizer.preprocess(input_string)
        self.assertEqual(expected, result)

    def test_should_replace_user_wrote_reply_second_type_sign_in_middle(self):
        replacement = '<tag>'
        input_string = 'This is test mircurt (03.03.16 20:39) писал(а): This is test'
        tokenizer = UserWroteRuReplacer(replace_tag_with=replacement)
        expected = 'This is test ' + replacement + ' This is test'

        result = tokenizer.preprocess(input_string)
        self.assertEqual(expected, result)

    def test_should_replace_user_wrote_reply_mixed_type(self):
        replacement = '<tag>'
        input_string = 'mircurt (03.03.16 20:39) писал(a) Вчера: This is test'
        tokenizer = UserWroteRuReplacer(replace_tag_with=replacement)
        expected = '' + replacement + ' This is test'

        result = tokenizer.preprocess(input_string)
        self.assertEqual(expected, result)

    def test_should_replace_accent_mark_with_nonword_replacer(self):
        replacement = ''
        input_string = 'в ́этом предл́ожении в́ыставлены удар́ения'
        tokenizer = NonWhitespaceAlphaNumPuncSpecSymbolsAllUnicodeReplacer(replace_tag_with=replacement)
        expected = 'в этом предложении выставлены ударения'
        result = tokenizer.preprocess(input_string)
        self.assertEqual(expected, result)


class StackingPreprocessorTest(unittest.TestCase):

    REGEX_1 = 'a'
    TARGET_1 = 'b'
    REGEX_2 = 'c'
    TARGET_2 = 'd'

    def test_should_replace_continuosly_without_collisions(self):
        replacer_1 = RegexReplacer(StackingPreprocessorTest.REGEX_1, StackingPreprocessorTest.TARGET_1, False, False, False)
        replacer_2 = RegexReplacer(StackingPreprocessorTest.REGEX_2, StackingPreprocessorTest.TARGET_2, False, False, False)

        stacking_replacer = StackingPreprocessor()
        stacking_replacer.append_preprocessor(replacer_1)
        stacking_replacer.append_preprocessor(replacer_2)

        input_string = 'acaca'
        output = stacking_replacer.preprocess(input_string)

        expected = 'bdbdb'
        self.assertEqual(expected, output)

    def test_should_replace_continuosly_with_chaining(self):
        replacer_1 = RegexReplacer(StackingPreprocessorTest.REGEX_1, StackingPreprocessorTest.TARGET_1, False, False, False)
        replacer_2 = RegexReplacer(StackingPreprocessorTest.TARGET_1, StackingPreprocessorTest.TARGET_2, False, False, False)

        stacking_replacer = StackingPreprocessor()

        stacking_replacer.append_preprocessor(replacer_1)
        stacking_replacer.append_preprocessor(replacer_2)

        input_string = 'acaca'
        output = stacking_replacer.preprocess(input_string)

        expected = 'dcdcd'
        self.assertEqual(expected, output)

    def test_should_replace_in_right_order_when_prepend(self):
        replacer_1 = RegexReplacer(StackingPreprocessorTest.TARGET_2, StackingPreprocessorTest.TARGET_1, False, False, False)
        replacer_2 = RegexReplacer(StackingPreprocessorTest.REGEX_2, StackingPreprocessorTest.TARGET_2, False, False, False)

        stacking_replacer = StackingPreprocessor()

        stacking_replacer.append_preprocessor(replacer_1)
        stacking_replacer.prepend_preprocessor(replacer_2)

        input_string = 'acaca'
        output = stacking_replacer.preprocess(input_string)

        expected = 'ababa'
        self.assertEqual(expected, output)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import io
import unittest

from pynlple.data.filesource import FilePathSource
from ..module import abs_path as path


class FileSourceTest(unittest.TestCase):

    def setUp(self):
        self.encoding = 'utf8'
        self.filepath1 = path(__file__, 'data/filesource/file1.txt')
        self.filepath2 = path(__file__, 'data/filesource/file2.txt')

    def test_should_two_readers_yield_lines_simultaneously(self):
        with io.open(self.filepath1, 'rt', encoding=self.encoding) as file1,\
                io.open(self.filepath2, 'rt', encoding=self.encoding) as file2:
            lines = [[next(file_) for file_ in [file1, file2]] for _ in range(4)]
            self.assertEqual(self.readlines(self.filepath1), [line[0] for line in lines])
            self.assertEqual(self.readlines(self.filepath2), [line[1] for line in lines])

    def readlines(self, filepath):
        with io.open(filepath, 'rt', encoding=self.encoding) as file:
            return [line for line in file]




# -*- coding: utf-8 -*-
from pynlple.module import abs_path
from pynlple.processing.dictionary import DictionaryLookUp, FileFolderTokenMapping
from pynlple.processing.text import ClassTokenFilter, TokenFilter, DictionaryBasedTagger
import re
from operator import contains
from functools import partial


class MustContainTokenFilter(TokenFilter):

    def __init__(self, all_substring_set=None, any_substring_set=None,):
        self.all_set = all_substring_set
        self.any_set = any_substring_set

    def filter(self, list_):
        return filter(self.__contains, list_)

    def __contains(self, string_):
        return any(map(partial(contains, self.any_set), string_)) if self.any_set else True and \
            all(map(partial(contains, self.all_set), string_)) if self.all_set else True


class DictionaryStopwordsFilter(ClassTokenFilter):

    def __init__(self, stopwords_provider, prep=None):
        self.__stopwords_provider = stopwords_provider
        self.__dictionary = DictionaryLookUp(self.__stopwords_provider, preprocessing_method=prep)
        self.__tagger = DictionaryBasedTagger(self.__dictionary)
        super().__init__(self.__tagger)


class PunctuationFilter(DictionaryStopwordsFilter):
    """
    Dictionary-based punctuation filter. Uses punctuation from `data/punctuation.txt`.

    Example usage::

    >>> filter_ = PunctuationFilter()
    >>> filter_.filter('Some , tokens ! and . punctuation !'.split())
    ['Some', 'tokens', 'and', 'punctuation']

    """

    def __init__(self, prep=None):
        self.__stopwords_provider = FileFolderTokenMapping(
            ['punctuation.txt'],
            data_folder=abs_path(__file__, 'data'),
            recursive=False)
        super().__init__(self.__stopwords_provider, prep)


class SpecialSymbolFilter(DictionaryStopwordsFilter):

    def __init__(self, prep=str.lower):
        self.__stopwords_provider = FileFolderTokenMapping(
            ['special_symbols.txt'],
            data_folder=abs_path(__file__, 'data'),
            recursive=False)
        super().__init__(self.__stopwords_provider, prep)


class BaseSymbolFilter(DictionaryStopwordsFilter):

    def __init__(self, prep=None):
        self.__stopwords_provider = FileFolderTokenMapping(
            ['punctuation.txt', 'special_symbols.txt'],
            data_folder=abs_path(__file__, 'data'),
            recursive=True)
        super().__init__(self.__stopwords_provider, prep)


class CurrencyFilter(DictionaryStopwordsFilter):

    def __init__(self, prep=str.lower):
        self.__stopwords_provider = FileFolderTokenMapping(
            ['currency.txt'],
            data_folder=abs_path(__file__, 'data'),
            recursive=False)
        super().__init__(self.__stopwords_provider, prep)


class SpecialTagFilter(DictionaryStopwordsFilter):

    def __init__(self, prep=str.lower):
        self.__stopwords_provider = FileFolderTokenMapping(
            ['special_tags.txt'],
            data_folder=abs_path(__file__, 'data'),
            recursive=False)
        super().__init__(self.__stopwords_provider, prep)


class SymbolFilter(DictionaryStopwordsFilter):

    def __init__(self, prep=None):
        self.__stopwords_provider = FileFolderTokenMapping(
            ['punctuation.txt', 'special_symbols.txt', 'currency.txt'],
            data_folder=abs_path(__file__, 'data'),
            recursive=True)
        super().__init__(self.__stopwords_provider, prep)


class DefaultRussianStopwordsFilter(DictionaryStopwordsFilter):

    def __init__(self, prep=None):
        self.__stopwords_provider = FileFolderTokenMapping(
            {'rus/pos': ['conjunctions.txt', 'particles.txt', 'prepositions.txt', 'pronouns.txt']},
            data_folder=abs_path(__file__, 'data'),
            recursive=True)
        super().__init__(self.__stopwords_provider, prep)


class RegexFilter(TokenFilter):

    def __init__(self, regex):
        self.__regex = regex
        self.__pattern = re.compile(self.__regex)

    def filter(self, tokens):
        return list(filter(lambda t: not self.__pattern.match(t), tokens))


class NumberTokenFilter(RegexFilter):

    def __init__(self, num_regex=r'^[\.]?[\d][\d\.,\-:]*$'):
        super().__init__(num_regex)


class LengthTokenFilter(TokenFilter):

    def __init__(self, min_length):
        self.min_length = min_length
        self.__b = self.min_length - 1

    def filter(self, tokens):
        return list(filter(lambda x: len(x) > self.__b, tokens))


"""
    This module implements custom classifiers of Sklearn's Pipeline
"""
from pynlple.processing.stopwords import FileFolderTokenMapping, DictionaryBasedTagger, DictionaryLookUp
from pynlple.ml.transformers import ToSeries, VocabularyFeature, SubCategoryGenerator, ThresholdRuleVectorClassifier

from sklearn.preprocessing import FunctionTransformer
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.pipeline import Pipeline
from sklearn.base import BaseEstimator, ClassifierMixin

import pandas as pd
import numpy as np
from itertools import chain


class LexiconClassifier(BaseEstimator, ClassifierMixin):

    def __init__(self, lexicon_path, rules, stop_rules=list(), overlap_entities=False, class_gramms=(1,3)):
        self.lexicon_path = lexicon_path
        self.default_value = 0
        self.class_value = 1
        self.stop_value = -1
        self.rules = rules
        self.stop_rules = stop_rules
        self.overlap_entities = overlap_entities
        self.class_gramms = class_gramms
        self.join_s = '/'
        self._tagger = DictionaryBasedTagger(DictionaryLookUp(FileFolderTokenMapping([],
                                                                                     data_folder=self.lexicon_path),
                                                              preprocessing_method=str.lower))

        self._to_series = ToSeries()
        self._extractor = VocabularyFeature(self._tagger, self.overlap_entities)
        self._transformer_pipe = Pipeline([
                ('category_extractor', SubCategoryGenerator(self.class_gramms, self.join_s)),
                ('count_vectorizer', CountVectorizer(lowercase=False, analyzer=chain.from_iterable,)),
                ('to_array', FunctionTransformer(func=lambda x: x.toarray(), accept_sparse=True)),
        ])
        self._classifier = ThresholdRuleVectorClassifier(self.default_value)

        stop_classes = np.full((len(self.stop_rules)), self.stop_value)
        classes = np.full((len(self.rules)), self.class_value)
        all_classes = np.hstack((classes, stop_classes))
        all_rules = pd.Series(self.rules).append(pd.Series(self.stop_rules))

        self._transformer_pipe = self._transformer_pipe.fit(all_rules)
        self._classifier = self._classifier.fit(self._transformer_pipe.transform(all_rules), all_classes)

        self.tagger_pipeline = Pipeline([
                ('to_series', self._to_series),
                ('extractor', self._extractor),
                ('transformer', self._transformer_pipe),
                ('classifier', self._classifier),
        ])

    def __check_prediction_array(self, pred):
        return np.all(pred != self.stop_value) and np.any(pred == self.class_value)

    def transform(self, X):
        return self.tagger_pipeline.predict(X)

    def predict(self, X):
        return np.apply_along_axis(lambda x: self.__check_prediction_array(x), axis=1, arr=self.transform(X))


"""
    This module implements custom transformations of Sklearn's Pipeline
"""
from pynlple.processing.text import WSTokenizer as StringWSTokenizer
from pynlple.processing.text import RuleTokenizer as StringRuleTokenizer

from pandas import DataFrame, Categorical
from pandas.api.types import CategoricalDtype
import numpy as np
from sklearn.preprocessing._function_transformer import _identity
from sklearn.feature_extraction import DictVectorizer
from sklearn.base import BaseEstimator, TransformerMixin, ClassifierMixin


class NgramGenerator(object):

    def __init__(self, ngram_size=(1,1), join_s=' '):
        self.__size = ngram_size
        if self.__size[0] > self.__size[1]:
            raise ValueError('')
        self.__joiner = join_s
        super().__init__()

    def __call__(self, seq):
        # handle token n-grams
        min_n, max_n = self.__size
        if max_n != 1:
            original_seq = seq
            seq = []
            n_original_tokens = len(original_seq)
            for n in range(min_n,
                            min(max_n + 1, n_original_tokens + 1)):
                for i in range(n_original_tokens - n + 1):
                    seq.append(self.__joiner.join(original_seq[i: i + n]))
        return seq


class SENgramGenerator(object):

    def __init__(self, ngram_size=(1,1), join_s=' ', s=None, e=None):
        self.__size = ngram_size
        if len(self.__size) <= 0 or len(self.__size) > 2:
            raise ValueError('Ngram_size cannot be of length {}. Please pass a tuple(min,max)'.format(str(len(self.__size))))
        self.__min_n, self.__max_n = self.__size
        if self.__size[0] > self.__size[1]:
            raise ValueError('Max ngram_size cannot be bigger than min ngram size.')
        if self.__min_n < 1:
            raise ValueError('Min ngram_size cannot be smaller than 1.')
        if self.__min_n < 1:
            raise ValueError('Max ngram_size cannot be smaller than 1.')
        self.__s = s
        self.__e = e
        self.__joiner = join_s

        if self.__max_n > 1:
            self._ngrams = self._manygramm
        else:
            self._ngrams = self._onegramm

        if self.__s is not None and self.__e is None:
            self._preprocess = self._start
        elif self.__e is not None and self.__s is None:
            self._preprocess = self._end
        elif self.__s is not None and self.__e is not None:
            self._preprocess = self._start_end
        else:
            self._preprocess = self._pass
        super().__init__()

    def __call__(self, seq):
        return self._ngrams(list(self._preprocess(seq)))

    def _onegramm(self, seq):
        return seq

    def _manygramm(self, seq):
        gramms = []
        seq_len = len(seq)
        for n in range(self.__min_n, min(self.__max_n + 1, seq_len + 1)):
            for i in range(seq_len - n + 1):
                gramms.append(self.__joiner.join(seq[i: i + n]))
        return gramms

    def _start(self, seq):
        yield self.__s
        yield from seq

    def _end(self, seq):
        yield from seq
        yield self.__e

    def _start_end(self, seq):
        yield self.__s
        yield from seq
        yield self.__e

    def _pass(self, seq):
        yield from seq


class ColumnEqualFeature(BaseEstimator, TransformerMixin):

    def __init__(self, col_selector, opp_col_selector):
        self.col_selector = col_selector
        self.opp_col_selector = opp_col_selector

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        return self.col_selector.transform(X) == self.opp_col_selector.transform(X)

    def get_params(self, deep=True):
        return {'col_selector': self.col_selector, 'opp_col_selector': self.opp_col_selector}


class VocabularyFeature(BaseEstimator, TransformerMixin):

    def __init__(self, entity_tagger, overlap_entities=False):
        self.__tagger = entity_tagger
        self.__overlap = overlap_entities

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        return X.apply(lambda x: [e[2] for e in self.__tagger.get_tagged_entities(x, self.__overlap)])


class SubCategoryGenerator(BaseEstimator, TransformerMixin):

    def __init__(self, ngram_size=(1,1), join_s='/'):
        self.__ngram_size = ngram_size
        self.__join_s = join_s
        self.__ngrams = NgramGenerator(ngram_size=self.__ngram_size, join_s=self.__join_s)

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        return X.apply(lambda x: [self.__ngrams(e.split(self.__join_s)) for e in x])


class ThresholdRuleVectorClassifier(BaseEstimator, ClassifierMixin):

    def __init__(self, default_class=0):
        self.default_class = default_class

    def fit(self, X, y):
        self.rules_ = X
        self.classes_ = y
        return self

    def transform(self, X):
        return X

    def __compare_each_row(self, row):
        answers = np.zeros((self.rules_.shape[0]), dtype=self.classes_.dtype)
        for i in np.arange(self.rules_.shape[0]):
            if np.all(row >= self.rules_[i]):
                answers[i] = self.classes_[i]
            else:
                answers[i] = self.default_class
        return answers

    def predict(self, X):
        return np.apply_along_axis(
            self.__compare_each_row,
            axis=1,
            arr=X
        )


class NumFeatureThresholdProbabilityPredictor(BaseEstimator, ClassifierMixin):

    def __init__(self, value_threshold=5, class_thres=0.5):
        self._min_value = value_threshold
        self._thres = class_thres
        self._step = self._thres / self._min_value
        super().__init__()

    def fit(self, X, y):
        self.classes_ = np.unique(y)
        return self

    def predict_proba(self, X):
        negative_probas = (X - self._min_value) * self._step
        conditional_negative_probas = np.where(X < self._min_value, -negative_probas, 0.0)
        res = np.vstack(((self._thres + conditional_negative_probas).T, (self._thres - conditional_negative_probas).T))
        return res.T

    def predict(self, X):
        return np.where(self.predict_proba(X)[:,0] < self._thres, *self.classes_.tolist())

    def get_params(self, deep=False):
        return {'class_thres': self._thres,
                'value_threshold': self._min_value}


class ThresholdClassifier(BaseEstimator, ClassifierMixin):
    def __init__(self, threshold=0.5):
        """Classify samples based on whether they are above of below `threshold`"""
        self.threshold = threshold

    def fit(self, X, y):
        self.classes_ = np.unique(y)
        return self

    def predict(self, X):
        # the implementation used here breaks ties differently
        # from the one used in RFs:
        #return self.classes_.take(np.argmax(X, axis=1), axis=0)
        return np.where(X[:, 0] > self.threshold, *self.classes_.tolist())

    def get_params(self, deep=False):
        return {'threshold': self.threshold}


class ItemSelector(BaseEstimator, TransformerMixin):

    def __init__(self, field_name):
        self.field = field_name

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        return X[self.field]

    def get_params(self, deep=False):
        return {'field_name': self.field}


class PandasSelector(BaseEstimator, TransformerMixin):

    def __init__(self, col_indexer=None, row_indexer=None):
        self.col_indexer = col_indexer
        self.row_indexer = row_indexer

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        if self.col_indexer and self.row_indexer:
            return X.loc[self.row_indexer, self.col_indexer]
        elif self.col_indexer and not self.row_indexer:
            return X.loc[:,self.col_indexer]
        elif not self.col_indexer and self.row_indexer:
            return X.loc[self.row_indexer]
        else:
            return X

    def get_params(self, deep=False):
        return {'col_indexer': self.col_indexer,
                'row_indexer': self.row_indexer}


class SelectorFunctionTransformer(BaseEstimator, TransformerMixin):

    def __init__(self, selectors, func=None, inverse_func=None,
                 kw_args=None, inv_kw_args=None):
        self.selectors = selectors
        self.func = func
        self.inverse_func = inverse_func
        self.kw_args = kw_args
        self.inv_kw_args = inv_kw_args

    def fit(self, X, y=None):
        self.selectors = [v.fit(X, y) for v in self.selectors]
        return self

    def transform(self, X, y=None):
        return self._transform(selected_features=[v.transform(X,y) for v in self.selectors],
                               kw_args=self.kw_args, func=self.func)

    def inverse_transform(self, X, y=None):
        return self._transform(selected_features=[v.inverse_transform(X,y) for v in self.selectors],
                               kw_args=self.inv_kw_args, func=self.func)

    def _transform(self, selected_features, kw_args=None, func=None):
        if func is None:
            func = _identity
        return func(*selected_features, **kw_args if kw_args else {})


class ApplyToSeries(BaseEstimator, TransformerMixin):

    def __init__(self, func, *args, **kwds):
        self.func = func
        self.args = args
        self.kwds = kwds
        super().__init__()

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        return X.apply(func=self.func, *self.args, **self.kwds)

    def get_params(self, deep=False):
        return {'func': self.func, 'args': self.args, 'kwds': self.kwds}


class ApplyToDataframe(BaseEstimator, TransformerMixin):

    def __init__(self, func, axis=0, *args, **kwds):
        self.func = func
        self.axis = axis
        self.args = args
        self.kwds = kwds
        super().__init__()

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        return X.apply(func=self.func, axis=self.axis, *self.args, **self.kwds)

    def get_params(self, deep=False):
        return {'func': self.func, 'axis': self.axis, 'args': self.args, 'kwds': self.kwds}


class ApplyToPandas(BaseEstimator, TransformerMixin):

    def __init__(self, func, axis=None, args=(), **kwds):
        self.func = func
        self.axis = axis
        self.args = args
        self.kwds = kwds
        super().__init__()

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        if self.kwds:
            if self.axis:
                return X.apply(func=self.func, axis=self.axis, args=self.args, **self.kwds)
            else:
                return X.apply(func=self.func, args=self.args, **self.kwds)
        else:
            if self.axis:
                return X.apply(func=self.func, axis=self.axis, args=self.args,)
            else:
                return X.apply(func=self.func, args=self.args)

    def get_params(self, deep=False):
        return {'func': self.func, 'axis': self.axis, 'args': self.args, 'kwds': self.kwds}


class ReshapePandas(BaseEstimator, TransformerMixin):

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        return X.values.reshape((-1, 1))


class ReshapeNumpy(TransformerMixin):

    def __init__(self, shape):
        self.shape = shape

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        return X.reshape(self.shape)

    def get_params(self, deep=False):
        return {'shape': self.shape}


class LengthFeature(BaseEstimator, TransformerMixin):

    def __init__(self):
        super().__init__()

    def fit(self, X, y=None):
        return self

    def __length(self, item):
        return len(item)

    def transform(self, X):
        return np.reshape(a=np.array(X.apply(self.__length)), newshape=(-1,1))


class Preprocessor(BaseEstimator, TransformerMixin):

    def __init__(self, preprocessors):
        if isinstance(preprocessors, str):
            raise ValueError('Type {} is invalid for Preprocessor pipeline step. Must be some IPreprocessor.'.format(type(preprocessors)))
        try:
            self.__preps = list(preprocessors)
        except TypeError: # catch when for loop fails
            self.__preps = [preprocessors]

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        for prep in self.__preps:
            X = X.apply(prep.preprocess)
        return X

    def get_params(self, deep=False):
        return {'preprocessors': self.__preps}


class ListPreprocessor(BaseEstimator, TransformerMixin):

    def __init__(self, preprocessors):
        if isinstance(preprocessors, str):
            raise ValueError('Type {} is invalid for ListPreprocessor pipeline step. Must be some IPreprocessor.'.format(type(preprocessors)))
        try:
            self.__preps = list(preprocessors)
        except TypeError: # catch when for loop fails
            self.__preps = [preprocessors]

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        for prep in self.__preps:
            X = X.apply(lambda list_: list(map(prep.preprocess, list_)))
        return X

    def get_params(self, deep=False):
        return {'preprocessors': self.__preps}


class ToSeries(BaseEstimator, TransformerMixin):

    def __init__(self):
        super().__init__()

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        from pandas import Series
        import numpy as np
        if type(X) in (list, tuple, np.array):
            X = Series(X)
        return X

    def get_params(self, deep=False):
        return {}


class ToDataFrame(BaseEstimator, TransformerMixin):

    CATEGORY = 'category'

    def __init__(self, feature_name_map=None, feature_dtype_map=None, learn_categories=False):
        self.names = feature_name_map
        self.dtypes = feature_dtype_map
        self.learn_categories = learn_categories
        super().__init__()

    def __construct_name_list(self, X, y):
        names_list = np.arange(X.shape[1], dtype='O')
        if self.names:
            ids, names = zip(*self.names.items())
            names_list[list(ids)] = names
        self.names_list_ = names_list.tolist()

    def __construct_categories(self, X, y):
        if self.learn_categories and self.dtypes:
            self.categories_map_ = dict()
            for col_name, dtype in self.dtypes.items():
                if dtype == self.CATEGORY:
                    if col_name in self.names_list_:
                        col_index = self.names_list_.index(col_name)
                    else:
                        col_index = col_name
                    self.categories_map_[col_name] = CategoricalDtype(categories=np.unique(X[:,col_index]))

    def fit(self, X, y=None):
        if self.names:
            self.__construct_name_list(X, y)
        if self.learn_categories:
            self.__construct_categories(X, y)
        return self

    def transform(self, X):
        if self.names:
            X = DataFrame(data=X, columns=self.names_list_)
        else:
            X = DataFrame(data=X)
        if self.dtypes:
            for col_name, dtype in self.dtypes.items():
                if dtype == self.CATEGORY:
                    if self.learn_categories:
                        column = Categorical(X.loc[:,col_name].values.tolist(),
                                                dtype=self.categories_map_[col_name])

                    else:
                        column = Categorical(X.loc[:,col_name].values.tolist())
                else:
                    column = X.loc[:,col_name].astype(dtype)
                X.drop(labels=[col_name], axis=1, inplace=True)
                X.loc[:,col_name] = column
        return X

    def get_params(self, deep=False):
        return {'feature_name_map': self.names, 'feature_dtype_map': self.dtypes, 'learn_categories': self.learn_categories}


class PassBy(TransformerMixin):

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        return X


class CountFeature(TransformerMixin):

    def __init__(self, match_func):
        self.func = match_func
        super().__init__()

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        return X.apply(self.count_matches).values.reshape((-1,1))

    def count_matches(self, sequence):
        return sum(map(self.func, sequence))


class RelativeCountFeature(TransformerMixin):

    def __init__(self, match_func):
        self.func = match_func
        super().__init__()

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        return X.apply(self.relative_count).values.reshape((-1,1))

    def relative_count(self, sequence):
        len_ = len(sequence)
        if len_ > 0:
            return self.count_matches(sequence) / len_
        else:
            return 0.0

    def count_matches(self, sequence):
        return sum(self.func(s) for s in sequence)


class Tokenizer(BaseEstimator, TransformerMixin):

    def __init__(self, tokenizer):
        self.__tokenizer = tokenizer

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        return X.apply(self.__tokenizer.tokenize)

    def get_params(self, deep=True):
        return {'tokenizer': self.__tokenizer}


class WSTokenizer(Tokenizer):

    def __init__(self):
        super().__init__(StringWSTokenizer())


class PipeRuleTokenizer(Tokenizer):

    def __init__(self):
        super().__init__(StringRuleTokenizer())


class SpaceDetokenizer(BaseEstimator, TransformerMixin):

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        return X.apply(' '.join)


class TokenFilterer(BaseEstimator, TransformerMixin):

    def __init__(self, filters):
        if isinstance(filters, str):
            raise ValueError('Type {} is invalid for TokenFilterer pipeline step. Must be some pynlple filter.'.format(type(filters)))
        try:
            self.__filters = list(filters)
        except TypeError: # catch when for loop fails
            self.__filters = [filters]

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        for filt in self.__filters:
            X = X.apply(filt.filter)
        return X

    def get_params(self, deep=False):
        return {'filters': self.__filters}


class TextStatsVectorizer(BaseEstimator, TransformerMixin):
    """Extract features from each document with DictVectorizer"""

    def __init__(self):
        self.vectorizer = DictVectorizer(sparse=False, sort=False)

    def fit(self, X, y=None):
        return self

    def get_feature_names(self):
        return self.vectorizer.feature_names_
    
    def __count_text_stats(self, tokens):
        return {
            'contains_exclamation_point': '!' in tokens,
            'contains_url_tag': 'urltag' in tokens,
            'contains_reference': 'atreftag' in tokens,
            'contains_email': 'emailtag' in tokens,
            'contains_digit': '0' in tokens,
        }

    def transform(self, X):
        return self.vectorizer.fit_transform(X.apply(self.__count_text_stats))


class POSVectorizer(BaseEstimator, TransformerMixin):
    """Extract pos as features from each document with DictVectorizer"""

    def __init__(self):
        import pymorphy2
        self.vectorizer = DictVectorizer(sparse=False, sort=False)
        self.morph = pymorphy2.MorphAnalyzer()
        self.PARTS_OF_SPEECH = frozenset([
            'NOUN',  # имя существительное
            'ADJF',  # имя прилагательное (полное)
            'ADJS',  # имя прилагательное (краткое)
            'COMP',  # компаратив
            'VERB',  # глагол (личная форма)
            'INFN',  # глагол (инфинитив)
            'PRTF',  # причастие (полное)
            'PRTS',  # причастие (краткое)
            'GRND',  # деепричастие
            'NUMR',  # числительное
            'ADVB',  # наречие
            'NPRO',  # местоимение-существительное
            'PRED',  # предикатив
            'PREP',  # предлог
            'CONJ',  # союз
            'PRCL',  # частица
            'INTJ',  # междометие
        ])

    def fit(self, x, y=None):
        return self

    def get_feature_names(self):
        return self.vectorizer.feature_names_

    def __get_pos_count(self, tokens):
        rez = dict.fromkeys(self.PARTS_OF_SPEECH, 0)
        if not tokens or len(tokens) <= 0:
            return rez
        for word in tokens:
            pos = self.morph.parse(word)[0].tag.POS
            if pos:
                rez[pos] += 1
        return {k: v / len(list(tokens)) for k, v in rez.items()}

    def transform(self, X):
        return self.vectorizer.fit_transform(X.apply(self.__get_pos_count))



# -*- coding: utf-8 -*-
from ..utils import itemgetter


class MyStemAnalysis(object):

    def __init__(self, analysis_):
        self._analysis = analysis_

    def __iter__(self):
        return zip(self.texts(), self.analyses())

    def analyses(self):
        return map(itemgetter('analysis', []), self._analysis)

    def texts(self):
        return map(itemgetter('text', ''), self._analysis)



# -*- coding: utf-8 -*-
import re

from pynlple.utils import itemgetter

__rn_patt = re.compile(r'(\r|\n|&#10;|&#13;)+')
__b_patt = re.compile(r'</?b>')


def __find_local_index(parts, abs_index):
    cur_index = 0
    for i, part in enumerate(parts):
        part_len = len(part)
        upd_index = cur_index + part_len
        if upd_index > abs_index:
            return i, abs_index - cur_index
        cur_index = upd_index
    return len(parts), 0


def __cut_pos_analysis_title(mystem_pos_analysis):
    analysis = list(mystem_pos_analysis)
    rec_parts = list(map(itemgetter('text', ''), analysis))
    rec_text = ''.join(rec_parts)
    first_rn_match = __rn_patt.search(rec_text)
    if first_rn_match:
        part_cut_index, local_cut_index = __find_local_index(rec_parts, first_rn_match.start())
        updated_nodes = analysis[:part_cut_index + 1]
        updated_node = dict(updated_nodes[part_cut_index])
        updated_node['text'] = updated_node.get('text', '')[:local_cut_index]
        updated_nodes[part_cut_index] = updated_node
        return updated_nodes
    return analysis


def __cut_pos_analysis_description(mystem_pos_analysis):
    analysis = list(mystem_pos_analysis)
    rec_parts = list(map(itemgetter('text', ''), analysis))
    rec_text = ''.join(rec_parts)
    first_rn_match = __rn_patt.search(rec_text)
    if first_rn_match:
        part_cut_index, local_cut_index = __find_local_index(rec_parts, first_rn_match.end())
        updated_nodes = analysis[part_cut_index:]
        updated_node = dict(updated_nodes[0])
        updated_node['text'] = updated_node.get('text', '')[local_cut_index:]
        updated_nodes[0] = updated_node
        return updated_nodes
    return analysis


def cut_title_or_description(youtube_post_highlighted_text, pos_analysis):
    first_rn_match = __rn_patt.search(youtube_post_highlighted_text)
    if first_rn_match:
        b_matches = list(__b_patt.finditer(youtube_post_highlighted_text))
        if len(b_matches) > 0:
            f_b = b_matches[0]
            l_b = b_matches[-1]
            has_b_in_title = f_b.end() < first_rn_match.start()
            has_b_in_description = l_b.start() > first_rn_match.end()
            if has_b_in_title and not has_b_in_description:
                return youtube_post_highlighted_text[:first_rn_match.start()], __cut_pos_analysis_title(pos_analysis)
            elif has_b_in_description and not has_b_in_title:
                return youtube_post_highlighted_text[first_rn_match.end():], __cut_pos_analysis_description(pos_analysis)
    return youtube_post_highlighted_text, pos_analysis


def cut_text_with_highlights_if_present(youtube_post_raw_text, highlights):
    if len(highlights) > 0:
        first_rn_match = __rn_patt.search(youtube_post_raw_text)
        if first_rn_match:
            f_b = min(map(itemgetter('start'), highlights))
            l_b = max(map(itemgetter('end'), highlights))
            has_b_in_title = f_b < first_rn_match.start()
            has_b_in_description = l_b > first_rn_match.end()
            if has_b_in_title and not has_b_in_description:
                return youtube_post_raw_text[:first_rn_match.start()], __cut_highlights(highlights, first_rn_match.start(), first_rn_match.end())[0]
            elif has_b_in_description and not has_b_in_title:
                return youtube_post_raw_text[first_rn_match.end():], __cut_highlights(highlights, first_rn_match.start(), first_rn_match.end())[1]
    return youtube_post_raw_text, highlights


def __cut_highlights(highlights, cut_start, cut_end):
    l_highlights = []
    r_highlights = []

    for highlight in highlights:
        s = highlight['start']
        e = highlight['end']
        if s < cut_start and e <= cut_start:
            l_highlights.append(highlight)
        elif s > cut_end and e > cut_end:
            r_highlights.append(highlight)
        elif s < cut_start and e > cut_end:
            l_highlights.append({'start': s, 'end': cut_start})
            r_highlights.append({'start': cut_end, 'end': e})
        elif cut_start <= s <= cut_end and e > cut_end:
            r_highlights.append({'start': cut_end, 'end': e})
        elif s < cut_start and cut_start < e <= cut_end:
            l_highlights.append({'start': s, 'end': cut_start})

    return l_highlights, r_highlights



# -*- coding: utf-8 -*-
import re
from itertools import chain
from collections import defaultdict
from operator import itemgetter

from .dictionary import FileFolderTokenMapping
from .emojis import UnicodeEmojiProviderV11, UnicodeEmojiComponentProviderV11

TOKEN_DEFAULT_DELIM = ' '
PUNCTUATION = list(map(itemgetter(0), (map(itemgetter(0), FileFolderTokenMapping(['punctuation.txt'])))))
SPECIAL_SYMBOLS = list(map(itemgetter(0), (map(itemgetter(0), FileFolderTokenMapping(['special_symbols.txt'])))))
CURRENCY = list(map(itemgetter(0),map(itemgetter(0), FileFolderTokenMapping(['currency.txt']))))
EMOJIS = list(map(itemgetter(0), UnicodeEmojiProviderV11()))
EMOJI_COMPONENTS = defaultdict(lambda: list())
for component, class_ in UnicodeEmojiComponentProviderV11():
    EMOJI_COMPONENTS[class_].append(component)

SIMPLE_EMOJI_REGEX = '[' + r''.join(map(re.escape, EMOJI_COMPONENTS['main_component'])) + ']' \
                   + '[' + r''.join(map(re.escape, chain(EMOJI_COMPONENTS['tone'], EMOJI_COMPONENTS['modifier']))) + ']' + '*'
SIMPLE_COMPLEMENT_REGEX = '[' + r''.join(map(re.escape, EMOJI_COMPONENTS['complementary_component'])) + ']' \
                        + '[' + r''.join(map(re.escape, chain(EMOJI_COMPONENTS['tone'], EMOJI_COMPONENTS['modifier']))) + ']' + '*'
EMOJI_JOINER_REGEX = '[' + r''.join(map(re.escape, EMOJI_COMPONENTS['joiner'])) + ']'
EMOJI_LETTERS_REGEX = '[' + r''.join(map(re.escape, EMOJI_COMPONENTS['letter'])) + ']' + '+'
EMOJI_TONES_REGEX = '[' + r''.join(map(re.escape, EMOJI_COMPONENTS['tone'])) + ']'
EMOJI_KEYCAPS_REGEX = '[' + r''.join(map(re.escape,  EMOJI_COMPONENTS['keycapped_char'])) + ']' + '[' + r''.join(map(re.escape, EMOJI_COMPONENTS['modifier'])) + ']' + '+'
EMOJI_COMPOUND_REGEX = '(' + SIMPLE_EMOJI_REGEX + '(' + EMOJI_JOINER_REGEX + '(' + SIMPLE_COMPLEMENT_REGEX + ')' + ')' + '*' \
                     + '|' + EMOJI_LETTERS_REGEX \
                     + '|' + EMOJI_TONES_REGEX \
                     + '|' + EMOJI_KEYCAPS_REGEX + ')'

EMOJI_COMPOUND_EXTENDED_REGEX = '(' + SIMPLE_EMOJI_REGEX + '(' + EMOJI_JOINER_REGEX + '(' + SIMPLE_EMOJI_REGEX + ')' + ')' + '*' \
                     + '|' + EMOJI_LETTERS_REGEX \
                     + '|' + EMOJI_TONES_REGEX \
                     + '|' + EMOJI_KEYCAPS_REGEX + ')'

TOKEN_TYPES = [
            r'([^\W\d_]+([\'\-][^\W\d_]+)*[^\W\d_]*)',  # tokens, also with - and ' in middle
            r'([^\W_]+([\'\-][^\W\d_]+)+)',  # tokens, starting with number and ' - in middle ("7-fold", etc.)
            r'(\d+([,:.]\d+)*\d*)',  # digital sequences, also with , : . in middle
            r'(\.+)',
            r'(_)',  # underscore as being excluded by \w and previously by [^\W\d_]+
            r'(\r?\n)',
            EMOJI_COMPOUND_REGEX,
            r'([^\w\d\s])',  # all individual symbols that are not alphanum or whitespaces

        ]
DELIMITER_TYPES = [
    r'\d+(?=[^\W\d])',
    r'(?<=[^\W\d])\d+',

    r'(?<=^)[\'\-]',
    r'(?<=[\W_])[\'\-]',
    r'[\'\-](?=[\W\d_])',
    r'[\'\-](?=$)',
    r'(?<=\d)[\'\-](?=\d)',  # "'", "-" in "0-0"

    r'\.{2,}',

    r'(?<=^)[,:\.]',
    r'[,:\.](?=$)',

    r'(?<![\d])[,:\.]',  # not dd.dd - exclude , : . in the middle digital sequences
    r'[,:\.](?![\d])',


    r'\r?\n',
    r'|'.join(map(re.escape, EMOJIS)),

    r'[^\w\d\s\'\-,:\.]',
    r'_',
]
DELIMITER_CONTEXTS = [
    r'(\d+)[^\W\d]',
    r'[^\W\d](\d+)',

    r'^([\'\-])',
    r'[\W_]([\'\-])',
    r'([\'\-])[\W\d_]',
    r'([\'\-])$',

    r'\d([\'\-])\d',

    r'(\.{2,})',

    r'^([,:\.])',
    r'([,:\.])$',

    r'[^\d]([,:\.])',  # not dd.dd - exclude , : . in the middle digital sequences
    r'([,:\.])[^\d]',

    r'(\r?\n)',
    r'(' + '|'.join(map(re.escape, EMOJIS)) + ')',

    r'([^\w\d\s\'\-,:\.])',
    r'(_)',
]


def has_capital_letter(token):
    return any(l.isupper() for l in token)


def first_capital_letter(token):
    return len(token) > 0 and token[0].isupper()


def only_first_capital_letter(token):
    return first_capital_letter(token) and not has_capital_letter(token[1:])


def has_not_first_capital_letter(token):
    return not first_capital_letter(token) and has_capital_letter(token[1:])


def one_capital_letter(token):
    return sum(l.isupper() for l in token) == 1


def more_than_one_capital_letter(token):
    return sum(l.isupper() for l in token) > 1


def all_capital(token):
    return token.isupper()


def all_lower(token):
    return token.islower()


newline_pattern = re.compile(r'\r\n?')
def is_newline(token):
    return newline_pattern.fullmatch(token) is not None


multidot_pattern = re.compile(r'\.\.+')
def is_multiple_dot(token):
    return multidot_pattern.fullmatch(token) is not None


digital_sequence_pattern = re.compile(r'^[\.]?[\d][\d\.,\-:]*$')
def is_a_digital_sequence(token):
    return digital_sequence_pattern.fullmatch(token) is not None


# -*- coding: utf-8 -*-
import logging
import io
import gzip
import bz2

from gensim.models.doc2vec import TaggedDocument
from pynlple.data.source import Source


class StackingSource(Source):

    logger = logging.getLogger(__name__)

    def __init__(self, list_sources, log=False):
        self.sources = list_sources
        self.log = log

    def __iter__(self):
        # if self.log:
            # self.logger.info('[%s] Corpus iterator %s started yielding elements.', str(self.__class__.__name__), repr(self))
        for i_, source in enumerate(self.sources):
            if self.log:
                self.logger.info('[%s] Corpus iterator started yielding elements from source (%d/%d): %s.',
                             str(self.__class__.__name__), i_, len(self.sources), repr(source))
            for item in source:
                yield item


class DeduplicatingFixedCacheSizeSource(Source):

    def __init__(self, source, cache_size=10000, refresh=False, feature_extractor=lambda x: x):
        self.source = source
        self.cache_size = cache_size
        self.refresh = refresh
        self.feature_extractor = feature_extractor

    def __iter__(self):
        from collections import OrderedDict
        self.__cache = OrderedDict()
        stub = 0
        for entry in self.source:
            f_entry = self.feature_extractor(entry)
            if f_entry in self.__cache:
                if self.refresh:
                    self.__cache.move_to_end(f_entry)
            else:
                self.__cache.__setitem__(f_entry, stub)
                if self.cache_size is not None:
                    while len(self.__cache) > self.cache_size:
                        self.__cache.popitem(last=False)
                yield entry


class JsonFieldSource(Source):

    def __init__(self, json_source, key, default=None):
        self.json = json_source
        self.key = key
        self.default = default

    def __iter__(self):
        for json_entry in self.json:
            if self.key in json_entry:
                yield json_entry[self.key]
            else:
                yield self.default


class FilteringSource(Source):

    def __init__(self, source, condition):
        self.source = source
        self.condition = condition

    def __iter__(self):
        for entry in self.source:
            if self.condition(entry):
                yield entry


class MappingSource(Source):

    def __init__(self, source, function):
        self.source = source
        self.function = function

    def __iter__(self):
        for entry in self.source:
            yield self.function(entry)


class SplittingSource(Source):

    def __init__(self, source, splitting_function):
        self.source = source
        self.function = splitting_function

    def __iter__(self):
        for entry in self.source:
            for item in self.function(entry):
                yield item


class DFsTaggedDocumentSource(Source):

    def __init__(self, dataframe_sources, text_column, tag_columns=None):
        self.sources = dataframe_sources
        self.text_column = text_column
        self.tag_columns = tag_columns
        super().__init__()

    def __iter__(self):
        for source in self.sources:
            return DFTaggedDocumentSource(source, self.text_column, self.tag_columns)


class DFTaggedDocumentSource(Source):

    def __init__(self, dataframe_source, text_column, tag_columns=None):
        self.source = dataframe_source
        self.text_column = text_column
        self.tag_columns = tag_columns

    def __iter__(self):
        for row in self.source.get_dataframe().itertuples():
            text = getattr(row, self.text_column)
            tokens = text.split()
            if self.tag_columns:
                yield TaggedDocument(tokens, [getattr(row, tag_column) for tag_column in self.tag_columns])
            else:
                yield tokens


class FileLineSource(Source):

    def __init__(self, text_file_path, encoding='utf8'):
        self.source_file = text_file_path
        self.encoding = encoding

    def __iter__(self):
        with io.open(self.source_file, mode='rt', encoding=self.encoding) as in_file:
            for line in in_file:
                line = line.strip()
                if len(line) <= 0:
                    continue
                yield line


class OpensubtitlesSentenceSource(Source):

    DEFAULT_SENTENCE_TAG = '<s>'

    def __init__(self, line_source, sentence_tag=None):
        self.source = line_source
        if sentence_tag:
            self.sentence_tag = sentence_tag
        else:
            self.sentence_tag = OpensubtitlesSentenceSource.DEFAULT_SENTENCE_TAG

    def __iter__(self):
        for line in self.source:
            for sentence in line.split(self.sentence_tag):
                yield sentence.strip()


class BZipDocumentSource(Source):

    def __init__(self, bzip_filepath, text_preprocessor=None):
        self.source_filepath = bzip_filepath
        self.text_preprocessor = text_preprocessor
        super().__init__()

    def __iter__(self):
        with bz2.BZ2File(self.source_filepath, 'rtU') as in_bz:
            for line in in_bz:
                text = line
                if self.text_preprocessor:
                    text = self.text_preprocessor.preprocess(text)
                tokens = text.split()
                yield tokens


class GZipDocumentSource(Source):

    def __init__(self, gzip_filepath, text_preprocessor=None):
        self.source_filepath = gzip_filepath
        self.text_preprocessor = text_preprocessor
        super().__init__()

    def __iter__(self):
        with gzip.GzipFile(self.source_filepath, 'rU') as in_bz:
            for line in in_bz:
                text = line
                if self.text_preprocessor:
                    text = self.text_preprocessor.preprocess(text)
                tokens = text.split()
                yield tokens



import numpy as np
import pandas as pd
from sklearn.cross_validation import KFold, StratifiedKFold, cross_val_score
from sklearn import metrics


def cross_validation_training(data, pipeline, n_folds=6, verbose=False):
    # Partitioning a sample of data into complementary subsets for CrossValidation
    k_fold = KFold(n=len(data), n_folds=n_folds, shuffle=True)

    precision, recall = [], []
    confusion = np.array([[0, 0], [0, 0]])

    results = {
        'text': [], 'golden': [], 'predicted': [], 'p1': []
    }

    for train_indices, test_indices in k_fold:
        # Prepare training data
        train_text = data.iloc[train_indices]['Contents'].values
        train_y = data.iloc[train_indices]['Registered'].values

        # Prepare test data
        test_text = data.iloc[test_indices]['Contents'].values
        test_y = data.iloc[test_indices]['Registered'].values

        # Train the model
        pipeline.fit(train_text, train_y)

        # Make predictions
        in_training = set(train_text)
        old_t, old_y = [], []
        new_t, new_y = [], []
        for tt, ty in zip(test_text, test_y):
            if tt in in_training:
                old_t.append(tt)
                old_y.append(ty)
            else:
                new_t.append(tt)
                new_y.append(ty)
        if verbose:
            print('%d out of %d are from training' % (len(old_y), len(test_y)))
            tuples = [(np.array(new_t), np.array(new_y)),
                      (np.array(old_t), np.array(old_y))]
        else:
            tuples = [(np.array(new_t), np.array(new_y))]

        for i, (ttext, ty) in enumerate(tuples):
            predictions = pipeline.predict(ttext)
            prob = pipeline.predict_proba(ttext)
            if verbose:
                print(np.mean(predictions == ty))
            if i == 0:
                # Calculate precision and confusion matrix
                confusion += metrics.confusion_matrix(ty, predictions)
                precision.append(metrics.precision_score(ty, predictions))
                recall.append(metrics.recall_score(ty, predictions))

                results['text'] += ttext.tolist()
                results['golden'] += ty.tolist()
                results['predicted'] += predictions.tolist()
                results['p1'] += prob[:, 1].tolist()

    precision, recall = np.array(precision), np.array(recall)
    print("Precision: %0.3f (+/- %0.3f)" % (precision.mean(), precision.std() * 2))
    print("Recall: %0.3f (+/- %0.3f)" % (recall.mean(), recall.std() * 2))
    if verbose:
        print('Confusion matrix:')
        print(confusion)

    # Check predictions
    results = pd.DataFrame(results)
    return results
